# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serialization, serializers, and helper methods for going to and from flattened python dictionary
representations of events.

Attributes:
    JsonSerializer (Serializer): For serializing to and from json.

"""
import io
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union, Dict, Any, Optional

import mtap
from mtap import data, processing
from mtap import descriptions as d

logger = logging.getLogger(__name__)

__all__ = [
    'event_to_dict',
    'document_to_dict',
    'dict_to_event',
    'dict_to_document',
    'Serializer',
    'SerializationProcessor',
    'JsonSerializer',
    'YamlSerializer',
    'PickleSerializer',
    'standard_serializers',
    'get_serializer'
]


def event_to_dict(event: mtap.Event, *, include_label_text: bool = False) -> Dict:
    """A helper method that turns an event into a python dictionary.

    Args:
        event (Event): The event object.

    Keyword Args:
        include_label_text (bool): Whether to include the text labels cover with the labels.

    Returns:
        dict: A dictionary object suitable for serialization.

    """
    d = {
        'event_id': event.event_id,
        'metadata': {},
        'documents': {}
    }
    for k, v in event.metadata.items():
        d['metadata'][k] = v
    for doc in event.documents.values():
        d['documents'][doc.document_name] = document_to_dict(doc,
                                                             include_label_text=include_label_text)
    return d


def document_to_dict(document: mtap.Document, *, include_label_text: bool = False) -> Dict:
    """A helper method that turns a document into a python dictionary.

    Args:
        document (Document): The document object.

    Keyword Args:
        include_label_text (bool): Whether to include the text labels cover with the labels.

    Returns:
        dict: A dictionary object suitable for serialization.
    """
    d = {
        'text': document.text,
        'label_indices': {}
    }

    for index_name, index in document.labels.items():
        adapter = index.adapter
        if adapter is None:
            adapter = data.GENERIC_ADAPTER
        d['label_indices'][index_name] = adapter.pack(
            index,
            include_label_text=include_label_text
        )
    return d


def dict_to_event(d: Dict, *, client: Optional[mtap.EventsClient] = None) -> mtap.Event:
    """Turns a serialized dictionary into an Event.

    Args:
        d (dict): The dictionary representation of the event.
        client (~typing.Optional[EventsClient]): An events service to create the event on.

    Returns:
        Event: The deserialized event object.
    """
    event = mtap.Event(event_id=d['event_id'], client=client)
    for k, v in d['metadata'].items():
        event.metadata[k] = v
    for k, v in d['documents'].items():
        dict_to_document(k, v, event=event)
    return event


def dict_to_document(document_name: str,
                     d: Dict,
                     *, event: Optional[mtap.Event] = None) -> mtap.Document:
    """Turns a serialized dictionary into a Document.

    Args:
        document_name (str): The name identifier of the document on the event.
        d (dict): The dictionary representation of the document.
        event (~typing.Optional[Event]): An event that the document should be added to.

    Returns:
        Document: The deserialized Document object.

    """
    document = mtap.Document(document_name=document_name, text=d['text'])
    if event is not None:
        event.add_document(document)
    for k, v in d['label_indices'].items():
        adapter = document.get_default_adapter(k)
        index = adapter.unpack(v, k, document=document)
        document.add_labels(k, index, distinct=index.distinct)

    return document


class Serializer(ABC):
    """Abstract base class for a serializer of MTAP events.
    """

    @property
    @abstractmethod
    def extension(self) -> str:
        """str: Filename extension, including period. Ex: ``'.json'``."""
        ...

    @abstractmethod
    def event_to_file(self, event: mtap.Event, f: Union[Path, str, io.IOBase],
                      *, include_label_text: bool = False):
        """Writes the event to a file.

        Args:
            event (Event): The event object to serialize.
            f (~typing.Union[~pathlib.Path, str, ~io.IOBase]):
                A file or a path to a file to write the event to.
            include_label_text (bool):
                Whether, when serializing, to include the text that each label covers with the rest
                of the label.
        """
        ...

    @abstractmethod
    def file_to_event(self, f: Union[Path, str, io.IOBase], *,
                      client: Optional[mtap.EventsClient] = None) -> mtap.Event:
        """Loads an event from a serialized file.

        Args:
            f (~typing.Union[~pathlib.Path, str, ~io.IOBase]): The file to load from.
            client (~typing.Optional[EventsClient]): The events service to load the event into.

        Returns:
            Event: The loaded event object.
        """
        ...


@mtap.processor('mtap-serializer',
                description='Serializes events to a specific directory',
                parameters=[d.parameter('filename', data_type='str',
                                        description='Optional override for the filename to write the '
                                                    'document to.')])
class SerializationProcessor(mtap.EventProcessor):
    """An MTAP :obj:`EventProcessor` that serializes events to a specific directory.

    Args:
        ser (Serializer): The serializer to use.
        output_dir (str): The output_directory.
    """

    def __init__(self, ser: Serializer, output_dir: str, include_label_text: bool = False):
        self.serializer = ser
        self.output_dir = output_dir
        self.include_label_text = include_label_text
        Path(self.output_dir).mkdir(parents=True, exist_ok=True)

    def process(self, event: mtap.Event, params: Dict[str, Any]):
        name = params.get('filename', event.event_id + self.serializer.extension)
        path = Path(self.output_dir, name)
        self.serializer.event_to_file(event, path, include_label_text=self.include_label_text)


class _JsonSerializer(Serializer):
    """Serializer implementation that performs serialization to JSON.
    """

    @property
    def extension(self) -> str:
        return '.json'

    def event_to_file(self, event: mtap.Event, f: Path, *, include_label_text: bool = False):
        import json
        with processing.Processor.started_stopwatch('transform'):
            d = event_to_dict(event, include_label_text=include_label_text)
        with processing.Processor.started_stopwatch('io'):
            try:
                json.dump(d, f)
            except AttributeError:
                f = Path(f)
                f.parent.mkdir(parents=True, exist_ok=True)
                with f.open('w') as f:
                    json.dump(d, f)

    def file_to_event(self, f: Union[Path, str, io.IOBase],
                      client: Optional[mtap.EventsClient] = None) -> mtap.Event:
        import json
        with processing.Processor.started_stopwatch('io'):
            try:
                d = json.load(f)
            except AttributeError:
                if isinstance(f, str):
                    f = Path(f)
                with f.open('r') as f:
                    d = json.load(f)
        with processing.Processor.started_stopwatch('transform'):
            return dict_to_event(d, client=client)


class _YamlSerializer(Serializer):

    @property
    def extension(self) -> str:
        return '.yml'

    def event_to_file(self, event: mtap.Event, f: Union[Path, str, io.IOBase], *,
                      include_label_text: bool = False):
        import yaml
        try:
            from yaml import CDumper as Dumper
        except ImportError:
            from yaml import Dumper
        with processing.Processor.started_stopwatch('transform'):
            d = event_to_dict(event, include_label_text=include_label_text)
        with processing.Processor.started_stopwatch('io'):
            if isinstance(f, io.IOBase):
                yaml.dump(d, f, Dumper=Dumper)
            else:
                f = Path(f)
                with f.open('w') as f:
                    yaml.dump(d, f, Dumper=Dumper)

    def file_to_event(self, f: Union[Path, str, io.IOBase], *,
                      client: Optional[mtap.EventsClient] = None) -> mtap.Event:
        import yaml
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader
        with processing.Processor.started_stopwatch('io'):
            if isinstance(f, io.IOBase):
                d = yaml.load(f, Loader=Loader)
            else:
                with Path(f).open() as f:
                    d = yaml.load(f, Loader=Loader)
        with processing.Processor.started_stopwatch('transform'):
            return dict_to_event(d, client=client)


class _PickleSerializer(Serializer):

    @property
    def extension(self) -> str:
        return '.pickle'

    def event_to_file(self, event: mtap.Event, f: Union[Path, str, io.IOBase], *,
                      include_label_text: bool = False):
        import pickle
        with processing.Processor.started_stopwatch('transform'):
            d = event_to_dict(event, include_label_text=include_label_text)
        with processing.Processor.started_stopwatch('io'):
            try:
                pickle.dump(d, f)
            except TypeError:
                with Path(f).open('wb') as f:
                    pickle.dump(d, f)

    def file_to_event(self, f: Union[Path, str, io.IOBase], *,
                      client: Optional[mtap.EventsClient] = None) -> mtap.Event:
        import pickle
        with processing.Processor.started_stopwatch('io'):
            try:
                d = pickle.load(f)
            except TypeError:
                with Path(f).open('rb') as f:
                    d = pickle.load(f)
        with processing.Processor.started_stopwatch('transform'):
            return dict_to_event(d, client=client)


JsonSerializer = _JsonSerializer()
YamlSerializer = _YamlSerializer()
PickleSerializer = _PickleSerializer()

standard_serializers = {
    'json': JsonSerializer,
    'yml': YamlSerializer,
    'pickle': PickleSerializer
}


def get_serializer(identifier):
    return standard_serializers[identifier]



#  Copyright 2019 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from mtap import Event, label
from mtap.io.serialization import event_to_dict

text = "It seems that perfection is attained not when there is nothing more to add, but when " \
       "there is nothing more to remove."

tokens = [(0, 2), (3, 8), (9, 13), (14, 24), (25, 27), (28, 36), (37, 40), (41, 45), (46, 51),
          (52, 54), (55, 62), (63, 67), (68, 70), (71, 74), (76, 79), (80, 84), (85, 90), (91, 93),
          (94, 101), (102, 106), (107, 109), (110, 116)]


def test_event_to_dict_include_label_text():
    event = Event()
    doc = event.create_document('plaintext', text)
    doc.add_labels('sentences', [label(0, 117)])
    doc.add_labels('tokens', [label(start, end) for start, end in tokens])
    doc.add_labels('empty_index', [])

    d_event = event_to_dict(event, include_label_text=True)
    d_doc = d_event['documents']['plaintext']
    d_sentences = d_doc['label_indices']['sentences']
    assert d_sentences['labels'][0]['_text'] == text
    d_tokens = d_doc['label_indices']['tokens']['labels']
    for i, token in enumerate(d_tokens):
        assert token['_text'] == text[slice(*tokens[i])]
    assert d_doc['label_indices']['empty_index']['labels'] == []



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
from subprocess import PIPE, STDOUT, Popen, run

import pytest

from mtap.utilities import find_free_port, subprocess_events_server


@pytest.fixture(name='python_events')
def fixture_python_events():
    port = find_free_port()
    with subprocess_events_server(port=port) as address:
        yield address


@pytest.fixture(name='hello_processor')
def fixture_hello_processor(python_events, processor_watcher):
    port = find_free_port()
    p = Popen(['python', '-m', 'mtap.examples.tutorial.hello', '-p', str(port),
               '--events', python_events], stdin=PIPE, stdout=PIPE, stderr=STDOUT)
    address = "127.0.0.1:" + str(port)
    yield from processor_watcher(address=address, process=p)


@pytest.fixture(name='java_hello_processor')
def fixture_java_hello_processor(python_events, processor_watcher):
    mtap_jar = os.environ['MTAP_JAR']
    port = str(find_free_port())
    p = Popen(['java', '-cp', mtap_jar, 'edu.umn.nlpie.mtap.examples.HelloWorldExample',
               '-p', port, '--events', python_events],
              stdin=PIPE, stdout=PIPE, stderr=STDOUT)
    address = "127.0.0.1:" + port
    yield from processor_watcher(address=address, process=p)


@pytest.mark.integration
def test_hello_world(python_events, hello_processor):
    p = run(['python', '-m', 'mtap.examples.tutorial.pipeline', python_events, hello_processor],
            stdout=PIPE)
    p.check_returncode()
    assert p.stdout.decode('utf-8') == 'Hello YOUR NAME!\n'


@pytest.mark.integration
def test_java_hello_world(python_events, java_hello_processor):
    p = run(['python', '-m', 'mtap.examples.tutorial.pipeline', python_events,
             java_hello_processor], stdout=PIPE)
    p.check_returncode()
    assert p.stdout.decode('utf-8') == 'Hello YOUR NAME!\n'



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC, abstractmethod
from typing import Generic, Iterable, List, Any, Optional, Sequence, TypeVar

import typing

from mtap import _structs
from mtap.data import _label_indices, _labels

if typing.TYPE_CHECKING:
    import mtap
    from mtap.api.v1 import events_pb2
    from mtap import data

L = TypeVar('L', bound='data.Label')


class ProtoLabelAdapter(ABC, Generic[L]):
    """Responsible for marshalling and unmarshalling of label objects to and from proto messages.
    """
    __slots__ = ()

    @abstractmethod
    def create_label(self, *args, **kwargs) -> L:
        """Called by labelers to create labels.

        Should include the positional arguments `start_index` and `end_index`, because those are
        required properties of labels.

        Args:
            args: Arbitrary args used to create the label.
            kwargs: Arbitrary keyword args used to create the label.

        Returns:
            Label: An object of the label type.
        """
        ...

    @abstractmethod
    def create_index_from_response(
            self,
            response: 'events_pb2.GetLabelsResponse'
    ) -> 'data.LabelIndex[L]':
        """Creates a LabelIndex from the response from an events service.

        Args:
            response (mtap.api.v1.events_pb2.GetLabelsResponse): The response protobuf message from
                the events service.

        Returns:
            LabelIndex[L]: A label index containing all the labels from the events service.
        """
        ...

    @abstractmethod
    def create_index(self, labels: Iterable[L]):
        """Creates a LabelIndex from an iterable of label objects.

        Args:
            labels (~typing.Iterable[L]): Labels to put in index.

        Returns:
            LabelIndex[L]: A label index containing all of the labels in the list.
        """
        ...

    @abstractmethod
    def add_to_message(self, labels: List[L], request: 'events_pb2.AddLabelsRequest'):
        """Adds a list of labels to a request to the event service to add labels.

        Args:
            labels (Iterable[L]): The list of labels that need to be sent to the server.
            request (mtap.api.v1.events_pb2.AddLabelsRequest): The request proto message to add the
                labels to.
        """
        ...

    @abstractmethod
    def pack(self, index: 'data.LabelIndex[L]', *, include_label_text: bool = False) -> Any:
        """Prepares to serialize a label index by transforming the label index into a representation
        that can be dumped to json, yml, pickle, etc.

        Args:
            index: The index itself.
            include_label_text:
                Whether to include the label's text in the serialized representation
                (for informative reasons).

        Returns:
            A dictionary representation of the label index.

        """
        ...

    @abstractmethod
    def unpack(self,
               packed: Any,
               label_index_name: str,
               *, document: Optional['mtap.Document'] = None) -> 'data.LabelIndex[L]':
        """Takes a packed, serializable object and turns it into a full label index.

        Args:
            packed (Any): The packed representation.
            label_index_name (str): The index name of the label index.
            document (Document): The document this label index occurs on.

        Returns:
            LabelIndex[L]: The label index.

        """
        ...

    def store_references(self, labels: Sequence[L]):
        """Take all the references for the labels and turn them into static references.

        Args:
             labels (Sequence[L]): The labels to store the references on.
        """
        ...


class _GenericLabelAdapter(ProtoLabelAdapter['mtap.GenericLabel']):
    __slots__ = ('distinct',)

    def __init__(self, distinct):
        self.distinct = distinct

    def create_label(self, *args, **kwargs):
        return _labels.GenericLabel(*args, **kwargs)

    def create_index(self, labels: List[L]):
        return _label_indices.label_index(labels, self.distinct, adapter=self)

    def create_index_from_response(self, response):
        generic_labels = response.generic_labels
        labels = []
        for label_message in generic_labels.labels:
            fields = {}
            _structs.copy_struct_to_dict(label_message.fields, fields)
            reference_field_ids = {}
            _structs.copy_struct_to_dict(label_message.reference_ids, reference_field_ids)
            generic_label = _labels.GenericLabel(label_message.start_index, label_message.end_index,
                                                 identifier=label_message.identifier, fields=fields,
                                                 reference_field_ids=reference_field_ids)
            labels.append(generic_label)

        return _label_indices.presorted_label_index(labels, generic_labels.is_distinct, adapter=self)

    def add_to_message(self, labels: List['mtap.GenericLabel'], request):
        generic_labels = request.generic_labels
        for label in labels:
            label_message = generic_labels.labels.add()
            label_message.identifier = label.identifier
            label_message.start_index = label.start_index
            label_message.end_index = label.end_index
            _structs.copy_dict_to_struct(label.fields, label_message.fields, [label])
            _structs.copy_dict_to_struct(label.reference_field_ids, label_message.reference_ids,
                                         [label])

    def pack(self,
             index: 'data.LabelIndex[mtap.GenericLabel]',
             *, include_label_text: bool = False) -> Any:
        d = {
            'labels': [_label_to_dict(label, include_label_text) for label in index],
            'distinct': index.distinct
        }
        return d

    def unpack(self, packed: Any,
               label_index_name: str,
               *, document: Optional['mtap.Document'] = None) -> 'data.LabelIndex[L]':
        return _label_indices.label_index(
            [_dict_to_label(d, label_index_name, document) for d in packed['labels']],
            distinct=packed['distinct'], adapter=self
        )

    def store_references(self, labels: Sequence['mtap.GenericLabel']):
        for label in labels:
            for k, v in label.reference_cache.items():
                if k not in label.reference_field_ids:
                    label.reference_field_ids[k] = _convert_to_references(v)


def _convert_to_references(o):
    if o is None:
        return o
    if isinstance(o, _labels.Label):
        ref = '{}:{}'.format(o.label_index_name, o.identifier)
        return ref
    if isinstance(o, typing.Mapping):
        rep = {}
        for k, v in o.items():
            rep[k] = _convert_to_references(v)
        return rep
    if isinstance(o, Sequence):
        rep = [_convert_to_references(v) for v in o]
        return rep


def _label_to_dict(label, include_label_text):
    d = {'start_index': label.start_index,
         'end_index': label.end_index,
         'identifier': label.identifier,
         'fields': label.fields,
         'reference_ids': label.reference_field_ids}
    if include_label_text:
        d['_text'] = label.text
    return d


def _dict_to_label(d, label_index_name, document):
    return _labels.GenericLabel(d['start_index'], d['end_index'], identifier=d['identifier'],
                                fields=d['fields'], reference_field_ids=d['reference_ids'],
                                label_index_name=label_index_name, document=document)


GENERIC_ADAPTER = _GenericLabelAdapter(False)

DISTINCT_GENERIC_ADAPTER = _GenericLabelAdapter(True)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any

import mtap
from mtap.processing.descriptions import labels


@mtap.processor(
    'mtap-python-references-example',
    human_name='Python References Examples',
    description='Shows use of referential fields on labels.',
    outputs=[
        labels('referenced'),
        labels('map_references'),
        labels('list_references'),
        labels('references')
    ]
)
class ReferencesExampleProcessor(mtap.DocumentProcessor):
    def process_document(self, document: mtap.Document, params: Dict[str, Any]):
        referenced = [
            mtap.GenericLabel(0, 1),
            mtap.GenericLabel(1, 2),
            mtap.GenericLabel(2, 3),
            mtap.GenericLabel(3, 4)
        ]

        # references can be a map of strings to labels
        with document.get_labeler('map_references') as label_map_references:
            label_map_references(0, 4, ref={
                'a': referenced[0],
                'b': referenced[1],
                'c': referenced[2],
                'd': referenced[3]
            })

        # references can be a list of labels
        with document.get_labeler('list_references') as label_list_references:
            label_list_references(0, 2, ref=[referenced[0], referenced[1]])
            label_list_references(2, 3, ref=[referenced[2], referenced[3]])

        # references can be direct
        with document.get_labeler('references') as label_references:
            label_references(0, 2, a=referenced[0], b=referenced[1])
            label_references(2, 3, a=referenced[2], b=referenced[3])

        # referenced labels don't need to be added via "addLabels" or "Labeler.close" before label
        # indices that reference them.
        # The Document will delay uploading any label indices to the server until they are.
        document.add_labels('referenced', referenced)


if __name__ == '__main__':
    mtap.run_processor(ReferencesExampleProcessor())



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
from subprocess import Popen, PIPE, STDOUT

import pytest

from mtap import EventsClient, RemoteProcessor, Pipeline, Event, GenericLabel
from mtap.utilities import subprocess_events_server, find_free_port


@pytest.fixture(name='python_events')
def fixture_python_events():
    with subprocess_events_server(cwd=Path(__file__).parents[2]) as address:
        yield address


@pytest.fixture(name='python_references_processor')
def fixture_python_references_processor(python_events, processor_watcher):
    env = dict(os.environ)
    port = str(find_free_port())
    p = Popen(['python', '-m', 'mtap.examples.example_references_processor', '-p', port,
               '--events', python_events, '--log-level', 'DEBUG'], stdin=PIPE, stdout=PIPE, stderr=STDOUT, env=env)
    yield from processor_watcher(address="127.0.0.1:" + port, process=p)


@pytest.fixture(name="java_references_processor")
def fixture_java_references_processor(python_events, processor_watcher):
    mtap_jar = os.environ['MTAP_JAR']
    mtap_jar = mtap_jar + ':' + str(Path(__file__).parents[1] / 'slf4j-simple-1.7.30.jar')
    env = dict(os.environ)
    port = str(find_free_port())
    p = Popen(['java', '-cp', mtap_jar,
               'edu.umn.nlpie.mtap.examples.ReferenceLabelsExampleProcessor',
               '-p', port, '-e', python_events], stdin=PIPE, stdout=PIPE, stderr=STDOUT, env=env)
    yield from processor_watcher(address="127.0.0.1:" + port, process=p)


@pytest.mark.integration
def test_java_references(python_events, java_references_processor):
    with EventsClient(address=python_events) as client, Pipeline(
        RemoteProcessor('mtap-java-reference-labels-example-processor',
                        address=java_references_processor)
    ) as pipeline:
        with Event(event_id='1', client=client) as event:
            document = event.create_document('plaintext', 'abcd')
            pipeline.run(document)
            references = document.labels['references']
            assert references[0].a == GenericLabel(0, 1)
            assert references[0].b == GenericLabel(1, 2)
            assert references[1].a == GenericLabel(2, 3)
            assert references[1].b == GenericLabel(3, 4)

            map_references = document.labels['map_references']
            assert map_references[0].ref == {
                'a': GenericLabel(0, 1),
                'b': GenericLabel(1, 2),
                'c': GenericLabel(2, 3),
                'd': GenericLabel(3, 4)
            }

            list_references = document.labels['list_references']
            assert list_references[0].ref == [GenericLabel(0, 1), GenericLabel(1, 2)]
            assert list_references[1].ref == [GenericLabel(2, 3), GenericLabel(3, 4)]


@pytest.mark.integration
def test_python_references(python_events, python_references_processor):
    with EventsClient(address=python_events) as client, Pipeline(
            RemoteProcessor('mtap-python-references-example', address=python_references_processor)
    ) as pipeline:
        with Event(event_id='1', client=client) as event:
            document = event.create_document('plaintext', 'abcd')
            pipeline.run(document)
            references = document.labels['references']
            assert references[0].a == GenericLabel(0, 1)
            assert references[0].b == GenericLabel(1, 2)
            assert references[1].a == GenericLabel(2, 3)
            assert references[1].b == GenericLabel(3, 4)

            map_references = document.labels['map_references']
            assert map_references[0].ref == {
                'a': GenericLabel(0, 1),
                'b': GenericLabel(1, 2),
                'c': GenericLabel(2, 3),
                'd': GenericLabel(3, 4)
            }

            list_references = document.labels['list_references']
            assert list_references[0].ref == [GenericLabel(0, 1), GenericLabel(1, 2)]
            assert list_references[1].ref == [GenericLabel(2, 3), GenericLabel(3, 4)]



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from enum import Enum
from typing import NamedTuple


class LabelIndexType(Enum):
    """The type of serialized labels contained in the label index."""
    UNKNOWN = 0
    """Label index not set or type not known."""

    GENERIC = 1
    """JSON / Generic Label index"""

    CUSTOM = 2
    """Other / custom protobuf label index"""


LabelIndexType.UNKNOWN.__doc__ = """Label index not set or type not known."""
LabelIndexType.GENERIC.__doc__ = """JSON / Generic Label index"""
LabelIndexType.CUSTOM.__doc__ = """Other / custom protobuf label index"""


LabelIndexInfo = NamedTuple('LabelIndexInfo',
                            [('index_name', str),
                             ('type', LabelIndexType)])
LabelIndexInfo.__doc__ = """Information about a label index contained on a document."""
LabelIndexInfo.index_name.__doc__ = """str: The name of the label index."""
LabelIndexInfo.type.__doc__ = """LabelIndexType: The type of the label index."""



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from concurrent.futures.thread import ThreadPoolExecutor

import grpc_testing
import pytest
from grpc import StatusCode

from mtap import EventsClient
from mtap.api.v1 import events_pb2, events_pb2_grpc
from mtap.data import LabelIndexType, LabelIndexInfo


@pytest.fixture(name='events_channel')
def fixture_events_channel():
    yield grpc_testing.channel(
        [
            events_pb2.DESCRIPTOR.services_by_name['Events']
        ],
        grpc_testing.strict_real_time()
    )


@pytest.fixture(name='events_client_executor')
def fixture_events_client_executor():
    executor = ThreadPoolExecutor(max_workers=1)
    yield executor
    executor.shutdown(wait=True)


def test_get_label_index_info(events_channel, events_client_executor):
    def call():
        client = EventsClient(address='a', _pool=object(), _channel=events_channel)
        result = client.get_label_index_info(event_id='1', document_name='plaintext')
        return result

    future = events_client_executor.submit(call)

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetEventsInstanceId']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetEventsInstanceIdResponse()
    response.instance_id = '0'
    rpc.terminate(response, None, StatusCode.OK, '')

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabelIndicesInfo']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetLabelIndicesInfoResponse()
    first = response.label_index_infos.add()
    first.index_name = 'foo'
    first.type = events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.GENERIC
    second = response.label_index_infos.add()
    second.index_name = 'bar'
    second.type = events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.CUSTOM
    third = response.label_index_infos.add()
    third.index_name = 'baz'

    rpc.terminate(response, None, StatusCode.OK, '')

    infos = future.result(timeout=1)
    assert infos is not None
    assert len(infos) == 3
    assert infos[0].index_name == 'foo'
    assert infos[0].type == LabelIndexType.GENERIC
    assert infos[1].index_name == 'bar'
    assert infos[1].type == LabelIndexType.CUSTOM
    assert infos[2].index_name == 'baz'
    assert infos[2].type == LabelIndexType.UNKNOWN


def test_get_binary_data_names(events_channel, events_client_executor):
    def call():
        client = EventsClient(address='a', _pool=object(), _channel=events_channel)
        result = client.get_all_binary_data_names(event_id='1')
        return result

    future = events_client_executor.submit(call)

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetEventsInstanceId']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetEventsInstanceIdResponse()
    response.instance_id = '0'
    rpc.terminate(response, None, StatusCode.OK, '')

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetAllBinaryDataNames']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetAllBinaryDataNamesResponse(binary_data_names=['a', 'b', 'c'])
    rpc.terminate(response, None, StatusCode.OK, '')

    assert future.result(5) == ['a', 'b', 'c']


def test_add_binary_data(events_channel, events_client_executor):
    def call():
        client = EventsClient(address='a', _pool=object(), _channel=events_channel)
        result = client.add_binary_data(event_id='1', binary_data_name='foo', binary_data=b'\xFF')
        return result

    events_client_executor.submit(call)

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetEventsInstanceId']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetEventsInstanceIdResponse()
    response.instance_id = '0'
    rpc.terminate(response, None, StatusCode.OK, '')

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddBinaryData']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.AddBinaryDataResponse()
    rpc.terminate(response, None, StatusCode.OK, '')
    assert req.binary_data == b'\xFF'


def test_get_binary_data(events_channel, events_client_executor):
    def call():
        client = EventsClient(address='a', _pool=object(), _channel=events_channel)
        result = client.get_binary_data(event_id='1', binary_data_name='foo')
        return result

    future = events_client_executor.submit(call)

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetEventsInstanceId']
    )
    rpc.send_initial_metadata(())
    response = events_pb2.GetEventsInstanceIdResponse()
    response.instance_id = '0'
    rpc.terminate(response, None, StatusCode.OK, '')

    _, req, rpc = events_channel.take_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetBinaryData']
    )
    assert req.event_id == '1'
    assert req.binary_data_name == 'foo'
    rpc.send_initial_metadata(())
    response = events_pb2.GetBinaryDataResponse(binary_data=b'\xFF')
    rpc.terminate(response, None, StatusCode.OK, '')

    result = future.result(5)
    assert result == b'\xFF'



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Labels functionality."""
import threading
from abc import ABC, abstractmethod, ABCMeta
from queue import Queue
from typing import List, Tuple, Set, TYPE_CHECKING, TypeVar, NamedTuple, Any, Mapping, Sequence, \
    Union, Optional

if TYPE_CHECKING:
    import mtap
    from mtap import data


class Location(NamedTuple('Location', [('start_index', float), ('end_index', float)])):
    """A location in text, a tuple of (`start_index`, `end_index`).

    Used to perform comparison of labels based on their locations.

    Args:
        start_index (float):
            The start index inclusive of the location in text.
        end_index (float):
            The end index exclusive of the location in text.

    Attributes:
        start_index (float):
            The start index inclusive of the location in text.
        end_index (float):
            The end index exclusive of the location in text.
    """
    __slots__ = ()

    def covers(self, other: Union['data.Location', 'data.Label']):
        """Whether the span of text covered by this label completely overlaps the span of text
        covered by the ``other`` label or location.

        Args:
            other (~typing.Union[Location, Label]): A location or label to compare against.

        Returns:
            bool: ``True`` if `other` is completely overlapped/covered ``False`` otherwise.
        """
        return self.start_index <= other.start_index and self.end_index >= other.end_index

    def relative_to(self, location: Union['data.Location', 'data.Label', int]) -> 'data.Location':
        """Creates a location relative to the the same origin as ``location`` and makes it relative
        to ``location``.

        Args:
            location (int or Location or Label): A location to relativize this location to.

        Returns:
            ~data.Location: A copy with updated indices.

        Examples:
            >>> sentence = Location(10, 20)
            >>> token = Location(10, 15)
            >>> token.relative_to(sentence)
            Location(start_index=0, end_index=5)

        """
        try:
            start_index = location.start_index
        except AttributeError:
            start_index = location
        if not isinstance(start_index, int):
            raise ValueError('location must be Label, Location, or an int value')
        return Location(self.start_index - start_index, self.end_index - start_index)

    def offset_by(self, location: Union['data.Location', 'data.Label', int]) -> 'data.Location':
        """Creates a location by offsetting this location by an integer or the ``start_index`` of a
        location / label. Derelativizes this location.

        Args:
            location (int or Location or Label): A location to offset this location by.

        Returns:
            ~data.Location: A copy with updated indices.

        Examples:
            >>> sentence = Location(10, 20)
            >>> token_in_sentence = Location(0, 5)
            >>> token_in_sentence.offset_by(sentence)
            Location(start_index=10, end_index=15)

        """
        try:
            start_index = location.start_index
        except AttributeError:
            start_index = location
        if not isinstance(start_index, int):
            raise ValueError('location must be Label, Location, or an int value')
        return Location(self.start_index + start_index, self.end_index + start_index)


class Label(ABC, metaclass=ABCMeta):
    """An abstract base class for a label of attributes on text.
    """

    @property
    @abstractmethod
    def document(self) -> 'mtap.Document':
        """Document: The parent document this label appears on."""
        ...

    @document.setter
    @abstractmethod
    def document(self, value: 'mtap.Document'):
        """Sets the label's document, this will automatically be done when the label is created
        via a Document (i.e. get_label_index) or added to a document (i.e. via labeler or add_labels).
        """
        ...

    @property
    @abstractmethod
    def label_index_name(self) -> str:
        """str: The label index this label appears on."""
        ...

    @label_index_name.setter
    @abstractmethod
    def label_index_name(self, value: str):
        """Sets the name for the label index this label appears on. Will automatically be called
        when a label is added to a document via labeler or add_labels."""
        ...

    @property
    @abstractmethod
    def identifier(self) -> int:
        """int: The index of the label within its label index."""
        ...

    @identifier.setter
    @abstractmethod
    def identifier(self, value: int):
        """The index of the label within its label index. Labels will automatically be assigned
        this when added to a document via labeler or add_labels."""
        ...

    @property
    @abstractmethod
    def start_index(self) -> int:
        """int: The index of the first character of the text covered by this label.
        """
        ...

    @start_index.setter
    @abstractmethod
    def start_index(self, value: int):
        ...

    @property
    @abstractmethod
    def end_index(self) -> int:
        """int: The index after the last character of the text covered by this label.
        """
        ...

    @end_index.setter
    @abstractmethod
    def end_index(self, value: int):
        ...

    @property
    def location(self) -> Location:
        """Location: A tuple of (start_index, end_index) used to perform sorting and
            comparison first based on start_index, then based on end_index.
        """
        return Location(self.start_index, self.end_index)

    @property
    def text(self):
        """str: The slice of document text covered by this label. Will retrieve from events server
        if it is not cached locally.
        """
        return self.document.text[self.start_index:self.end_index]

    @abstractmethod
    def shallow_fields_equal(self, other) -> bool:
        """Tests if the fields on this label and locations of references are the same as another
        label.

        Args:
            other: The other label to test.

        Returns:
            True if all of the fields are equal and the references are at the same locations.

        """
        pass

    @abstractmethod
    def collect_floating_references(self, s):
        pass


L = TypeVar('L', bound=Label)

_repr_local = threading.local()


class GenericLabel(Label):
    """Default implementation of the Label class which uses a dictionary to store attributes.

    Will be suitable for the majority of use cases for labels.

    Args:
        start_index (int): The index of the first character in text to be included in the label.
        end_index (int): The index after the last character in text to be included in the label.

    Keyword Args:
        document (~typing.Optional[Document]): The parent document of the label. This will be
            automatically set if a the label is created via labeler.
        **kwargs : Arbitrary, any other fields that should be added to the label, values must be
            json-serializable.

    Examples:
        >>> pos_tag = pos_tag_labeler(0, 5)
        >>> pos_tag.tag = 'NNS'
        >>> pos_tag.tag
        'NNS'

        >>> pos_tag2 = pos_tag_labeler(6, 10, tag='VB')
        >>> pos_tag2.tag
        'VB'
    """

    def __init__(self, start_index: int, end_index: int, *,
                 identifier: Optional[int] = None,
                 document: Optional['mtap.Document'] = None,
                 label_index_name: Optional['str'] = None,
                 fields: Optional[dict] = None,
                 reference_field_ids: Optional[dict] = None,
                 **kwargs):
        self._document = document
        self._label_index_name = label_index_name
        self._identifier = identifier
        self._start_index = int(start_index)
        self._end_index = int(end_index)
        if fields is None:
            self.fields = {}
        else:
            self.fields = fields
        if reference_field_ids is None:
            self.reference_field_ids = {}
        else:
            self.reference_field_ids = reference_field_ids
        self.reference_cache = {}
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def document(self) -> 'mtap.Document':
        return self._document

    @document.setter
    def document(self, document: 'mtap.Document'):
        self._document = document

    @property
    def label_index_name(self) -> str:
        return self._label_index_name

    @label_index_name.setter
    def label_index_name(self, value: str):
        self._label_index_name = value

    @property
    def identifier(self) -> int:
        return self._identifier

    @identifier.setter
    def identifier(self, value: int):
        self._identifier = value

    @property
    def start_index(self) -> int:
        return self._start_index

    @start_index.setter
    def start_index(self, start_index: int):
        self._start_index = start_index

    @property
    def end_index(self) -> int:
        return self._end_index

    @end_index.setter
    def end_index(self, end_index: int):
        self._end_index = end_index

    def _is_reserved(self, key):
        return key in self.__dict__.keys() or key in vars(GenericLabel) or key in vars(Label)

    def __getattr__(self, item):
        try:
            return self.fields[item]
        except KeyError:
            pass
        try:
            return self.reference_cache[item]
        except KeyError:
            pass
        try:
            ref_value = self.reference_field_ids[item]
            self.reference_cache[item] = _dereference(ref_value, self.document)
            return self.reference_cache[item]
        except KeyError:
            raise AttributeError('Key "{}" not in fields, reference cache, or reference ids.'
                                 .format(item))

    def __setattr__(self, key, value):
        if key in ('document', 'label_index_name', 'identifier', 'start_index', 'end_index',
                   '_document', '_label_index_name', '_identifier', '_start_index', '_end_index',
                   'fields', 'reference_field_ids', 'reference_cache'):
            object.__setattr__(self, key, value)
            return
        if self._is_reserved(key):
            raise ValueError('The key "{}" is a reserved key.'.format(key))
        is_ref = _is_referential(value, [id(self)])
        if is_ref:
            self.reference_cache[key] = value
        else:
            self.fields[key] = value

    def __eq__(self, other):
        if not isinstance(other, GenericLabel):
            return False
        if other is self:
            return True
        if not self.location == other.location:
            return False
        return self.shallow_fields_equal(other)

    def shallow_fields_equal(self, other):
        if not self.fields == other.fields:
            return False
        refs = set(self.reference_field_ids.keys()).union(self.reference_cache.keys())
        other_refs = set(other.reference_field_ids.keys()).union(other.reference_cache.keys())
        if not refs == other_refs:
            return False
        for k in refs:
            try:
                if self.reference_field_ids[k] == other.reference_field_ids[k]:
                    continue
            except KeyError:
                pass
            self_k = getattr(self, k)
            other_k = getattr(other, k)
            if not _collect_locations(self_k) == _collect_locations(other_k):
                return False

        return True

    def __repr__(self):
        try:
            stack = _repr_local.stack
        except AttributeError:
            stack = set()
            _repr_local.stack = stack

        if id(self) in stack:
            return 'GenericLabel(...)'
        stack.add(id(self))
        attributes = [repr(self.start_index), repr(self.end_index)]
        for k, v in self.fields.items():
            attributes.append("{}={}".format(k, repr(v)))
        for k, v in self.reference_cache.items():
            attributes.append("{}={}".format(k, repr(v)))
        for k, v in self.reference_field_ids.items():
            if k not in self.reference_cache:
                attributes.append("{}=ref:{}".format(k, repr(v)))
        stack.remove(id(self))
        return "GenericLabel(".format() + ", ".join(attributes) + ")"

    def collect_floating_references(self, s):
        queue = Queue()
        for k, v in self.reference_cache.items():
            if v is not None:
                queue.put(v)
        while not queue.empty():
            o = queue.get_nowait()
            if isinstance(o, Label):
                if o.identifier is None:
                    s.add(id(o))
            elif isinstance(o, Mapping):
                for _, v in o.items():
                    if v is not None:
                        queue.put(v)
            elif isinstance(o, Sequence):
                for v in o:
                    if v is not None:
                        queue.put(v)


def label(start_index: int,
          end_index: int,
          *, document: Optional['mtap.Document'] = None,
          **kwargs) -> GenericLabel:
    """An alias for :class:`GenericLabel`.

    Args:
        start_index (int): The index of the first character in text to be included in the label.
        end_index (int): The index after the last character in text to be included in the label.
        document (~typing.Optional[Document]): The parent document of the label. This will be
            automatically set if a the label is created via labeler.
        **kwargs : Arbitrary, any other fields that should be added to the label, values must be
            json-serializable.

    """
    return GenericLabel(start_index, end_index, document=document, **kwargs)


def _staticize(labels: Sequence['data.Label'],
               document: 'mtap.Document',
               label_index_name: str) -> Tuple[List['data.Label'], Set[int]]:
    """Prepares a label index for serialization by finalizing sort order and setting label
    identifiers.

    Args:
        labels (~typing.Sequence[GenericLabel]): The labels in a label index.

    Returns:
        List['GenericLabel']: The labels sorted by position.
        Set[int]: A set of labels which a referenced by labels in this index.

    """
    labels = sorted(labels, key=lambda x: x.location)
    waiting_on = set()
    for i, l in enumerate(labels):
        l.document = document
        l.identifier = i
        l.label_index_name = label_index_name
    for l in labels:
        l.collect_floating_references(waiting_on)
    return labels, waiting_on


def _is_referential(o: Any, parents=None) -> bool:
    if parents is None:
        parents = [id(o)]
    if isinstance(o, (str, float, bool, int)) or o is None:
        return False
    elif isinstance(o, Label):
        return True
    elif isinstance(o, Mapping):
        map_is_ref = None
        for v in o.values():
            if id(v) in parents:
                raise ValueError('Recursive loop')
            x = _is_referential(v, parents + [id(v)])
            if map_is_ref is None:
                map_is_ref = x
            elif x != map_is_ref:
                raise TypeError('Label dictionaries cannot have mixes of references to labels'
                                'and primitive types.')
        return map_is_ref
    elif isinstance(o, Sequence):
        seq_is_ref = None
        for v in o:
            if id(v) in parents:
                raise ValueError('Recursive loop')
            x = _is_referential(v, parents + [id(v)])
            if seq_is_ref is None:
                seq_is_ref = x
            elif x != seq_is_ref:
                raise TypeError('Label lists cannot have mixes of references to labels'
                                'and primitive types.')
        return seq_is_ref
    else:
        raise TypeError('Unrecognized type')


def _dereference(o: Any, document: 'mtap.Document') -> Any:
    if o is None:
        return o
    if isinstance(o, str):
        label_index_name, label_id = o.split(':')
        label_index = document.labels[label_index_name]
        label_ = label_index[int(label_id)]
        return label_
    if isinstance(o, Mapping):
        replacement = {}
        for k, v in o.items():
            replacement[k] = _dereference(v, document)
        return replacement
    if isinstance(o, Sequence):
        replacement = [_dereference(v, document) for v in o]
        return replacement


def _collect_locations(o):
    if o is None:
        return None
    if isinstance(o, Label):
        return o.location
    if isinstance(o, Mapping):
        return {k: _collect_locations(v) for k, v in o.items()}
    if isinstance(o, Sequence):
        return [_collect_locations(v) for v in o]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mtap/api/v1/processing.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cmtap/api/v1/processing.proto\x12\x0bmtap.api.v1\x1a\x1cgoogle/api/annotations.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1cgoogle/protobuf/struct.proto\"9\n\x0c\x43reatedIndex\x12\x15\n\rdocument_name\x18\x01 \x01(\t\x12\x12\n\nindex_name\x18\x02 \x01(\t\"\x84\x01\n\x0eProcessRequest\x12\x14\n\x0cprocessor_id\x18\x01 \x01(\t\x12\x10\n\x08\x65vent_id\x18\x02 \x01(\t\x12!\n\x19\x65vent_service_instance_id\x18\x04 \x01(\t\x12\'\n\x06params\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xff\x01\n\x0fProcessResponse\x12\x41\n\x0btiming_info\x18\x01 \x03(\x0b\x32,.mtap.api.v1.ProcessResponse.TimingInfoEntry\x12\'\n\x06result\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x32\n\x0f\x63reated_indices\x18\x03 \x03(\x0b\x32\x19.mtap.api.v1.CreatedIndex\x1aL\n\x0fTimingInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration:\x02\x38\x01\"&\n\x0eGetInfoRequest\x12\x14\n\x0cprocessor_id\x18\x01 \x01(\t\"<\n\x0fGetInfoResponse\x12)\n\x08metadata\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xd5\x01\n\nTimerStats\x12\'\n\x04mean\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12&\n\x03std\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration\x12&\n\x03max\x18\x03 \x01(\x0b\x32\x19.google.protobuf.Duration\x12&\n\x03min\x18\x04 \x01(\x0b\x32\x19.google.protobuf.Duration\x12&\n\x03sum\x18\x05 \x01(\x0b\x32\x19.google.protobuf.Duration\"\'\n\x0fGetStatsRequest\x12\x14\n\x0cprocessor_id\x18\x01 \x01(\t\"\xca\x01\n\x10GetStatsResponse\x12\x11\n\tprocessed\x18\x01 \x01(\x05\x12\x10\n\x08\x66\x61ilures\x18\x02 \x01(\x05\x12\x44\n\x0ctiming_stats\x18\x03 \x03(\x0b\x32..mtap.api.v1.GetStatsResponse.TimingStatsEntry\x1aK\n\x10TimingStatsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.mtap.api.v1.TimerStats:\x02\x38\x01\x32\xf7\x02\n\tProcessor\x12\x81\x01\n\x07Process\x12\x1b.mtap.api.v1.ProcessRequest\x1a\x1c.mtap.api.v1.ProcessResponse\";\x82\xd3\xe4\x93\x02\x35\"0/v1/processors/{processor_id}/process/{event_id}:\x01*\x12p\n\x07GetInfo\x12\x1b.mtap.api.v1.GetInfoRequest\x1a\x1c.mtap.api.v1.GetInfoResponse\"*\x82\xd3\xe4\x93\x02$\x12\"/v1/processors/{processor_id}/info\x12t\n\x08GetStats\x12\x1c.mtap.api.v1.GetStatsRequest\x1a\x1d.mtap.api.v1.GetStatsResponse\"+\x82\xd3\xe4\x93\x02%\x12#/v1/processors/{processor_id}/statsB\x1b\n\x19\x65\x64u.umn.nlpie.mtap.api.v1b\x06proto3')



_CREATEDINDEX = DESCRIPTOR.message_types_by_name['CreatedIndex']
_PROCESSREQUEST = DESCRIPTOR.message_types_by_name['ProcessRequest']
_PROCESSRESPONSE = DESCRIPTOR.message_types_by_name['ProcessResponse']
_PROCESSRESPONSE_TIMINGINFOENTRY = _PROCESSRESPONSE.nested_types_by_name['TimingInfoEntry']
_GETINFOREQUEST = DESCRIPTOR.message_types_by_name['GetInfoRequest']
_GETINFORESPONSE = DESCRIPTOR.message_types_by_name['GetInfoResponse']
_TIMERSTATS = DESCRIPTOR.message_types_by_name['TimerStats']
_GETSTATSREQUEST = DESCRIPTOR.message_types_by_name['GetStatsRequest']
_GETSTATSRESPONSE = DESCRIPTOR.message_types_by_name['GetStatsResponse']
_GETSTATSRESPONSE_TIMINGSTATSENTRY = _GETSTATSRESPONSE.nested_types_by_name['TimingStatsEntry']
CreatedIndex = _reflection.GeneratedProtocolMessageType('CreatedIndex', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDINDEX,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.CreatedIndex)
  })
_sym_db.RegisterMessage(CreatedIndex)

ProcessRequest = _reflection.GeneratedProtocolMessageType('ProcessRequest', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSREQUEST,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.ProcessRequest)
  })
_sym_db.RegisterMessage(ProcessRequest)

ProcessResponse = _reflection.GeneratedProtocolMessageType('ProcessResponse', (_message.Message,), {

  'TimingInfoEntry' : _reflection.GeneratedProtocolMessageType('TimingInfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROCESSRESPONSE_TIMINGINFOENTRY,
    '__module__' : 'mtap.api.v1.processing_pb2'
    # @@protoc_insertion_point(class_scope:mtap.api.v1.ProcessResponse.TimingInfoEntry)
    })
  ,
  'DESCRIPTOR' : _PROCESSRESPONSE,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.ProcessResponse)
  })
_sym_db.RegisterMessage(ProcessResponse)
_sym_db.RegisterMessage(ProcessResponse.TimingInfoEntry)

GetInfoRequest = _reflection.GeneratedProtocolMessageType('GetInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINFOREQUEST,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.GetInfoRequest)
  })
_sym_db.RegisterMessage(GetInfoRequest)

GetInfoResponse = _reflection.GeneratedProtocolMessageType('GetInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETINFORESPONSE,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.GetInfoResponse)
  })
_sym_db.RegisterMessage(GetInfoResponse)

TimerStats = _reflection.GeneratedProtocolMessageType('TimerStats', (_message.Message,), {
  'DESCRIPTOR' : _TIMERSTATS,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.TimerStats)
  })
_sym_db.RegisterMessage(TimerStats)

GetStatsRequest = _reflection.GeneratedProtocolMessageType('GetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSTATSREQUEST,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.GetStatsRequest)
  })
_sym_db.RegisterMessage(GetStatsRequest)

GetStatsResponse = _reflection.GeneratedProtocolMessageType('GetStatsResponse', (_message.Message,), {

  'TimingStatsEntry' : _reflection.GeneratedProtocolMessageType('TimingStatsEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETSTATSRESPONSE_TIMINGSTATSENTRY,
    '__module__' : 'mtap.api.v1.processing_pb2'
    # @@protoc_insertion_point(class_scope:mtap.api.v1.GetStatsResponse.TimingStatsEntry)
    })
  ,
  'DESCRIPTOR' : _GETSTATSRESPONSE,
  '__module__' : 'mtap.api.v1.processing_pb2'
  # @@protoc_insertion_point(class_scope:mtap.api.v1.GetStatsResponse)
  })
_sym_db.RegisterMessage(GetStatsResponse)
_sym_db.RegisterMessage(GetStatsResponse.TimingStatsEntry)

_PROCESSOR = DESCRIPTOR.services_by_name['Processor']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031edu.umn.nlpie.mtap.api.v1'
  _PROCESSRESPONSE_TIMINGINFOENTRY._options = None
  _PROCESSRESPONSE_TIMINGINFOENTRY._serialized_options = b'8\001'
  _GETSTATSRESPONSE_TIMINGSTATSENTRY._options = None
  _GETSTATSRESPONSE_TIMINGSTATSENTRY._serialized_options = b'8\001'
  _PROCESSOR.methods_by_name['Process']._options = None
  _PROCESSOR.methods_by_name['Process']._serialized_options = b'\202\323\344\223\0025\"0/v1/processors/{processor_id}/process/{event_id}:\001*'
  _PROCESSOR.methods_by_name['GetInfo']._options = None
  _PROCESSOR.methods_by_name['GetInfo']._serialized_options = b'\202\323\344\223\002$\022\"/v1/processors/{processor_id}/info'
  _PROCESSOR.methods_by_name['GetStats']._options = None
  _PROCESSOR.methods_by_name['GetStats']._serialized_options = b'\202\323\344\223\002%\022#/v1/processors/{processor_id}/stats'
  _CREATEDINDEX._serialized_start=137
  _CREATEDINDEX._serialized_end=194
  _PROCESSREQUEST._serialized_start=197
  _PROCESSREQUEST._serialized_end=329
  _PROCESSRESPONSE._serialized_start=332
  _PROCESSRESPONSE._serialized_end=587
  _PROCESSRESPONSE_TIMINGINFOENTRY._serialized_start=511
  _PROCESSRESPONSE_TIMINGINFOENTRY._serialized_end=587
  _GETINFOREQUEST._serialized_start=589
  _GETINFOREQUEST._serialized_end=627
  _GETINFORESPONSE._serialized_start=629
  _GETINFORESPONSE._serialized_end=689
  _TIMERSTATS._serialized_start=692
  _TIMERSTATS._serialized_end=905
  _GETSTATSREQUEST._serialized_start=907
  _GETSTATSREQUEST._serialized_end=946
  _GETSTATSRESPONSE._serialized_start=949
  _GETSTATSRESPONSE._serialized_end=1151
  _GETSTATSRESPONSE_TIMINGSTATSENTRY._serialized_start=1076
  _GETSTATSRESPONSE_TIMINGSTATSENTRY._serialized_end=1151
  _PROCESSOR._serialized_start=1154
  _PROCESSOR._serialized_end=1529
# @@protoc_insertion_point(module_scope)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MTAP SDK configuration."""
import logging
import os
import threading
from pathlib import Path
from typing import Any, MutableMapping, Iterator, Union

from mtap.utilities import mtap_home

logger = logging.getLogger(__name__)


def _collapse(d, path, v):
    d[path] = v
    try:
        p = ''
        if path is not None:
            p = path + '.'
        for k, v in v.items():
            _collapse(d, p + k, v)
        return d
    except AttributeError:
        pass
    except TypeError:
        raise ValueError('Failed to load configuration')
    return d


def _load_config(f):
    from yaml import load
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    config = load(f, Loader=Loader)
    if not isinstance(config, dict):
        raise TypeError("Failed to load configuration from file: " + str(f))
    return _collapse({}, None, config)


def _load_default_config():
    potential_paths = []
    try:
        cnf = os.getenv('MTAP_CONFIG')
        potential_paths.append(Path(cnf))
    except TypeError:
        pass
    locations = [Path.cwd(), mtap_home(), Path('/etc/mtap/')]
    potential_paths += [location / 'mtapConfig.yml' for location in locations]
    potential_paths.append(Path(__file__).parent / 'defaultConfig.yml')
    for config_path in potential_paths:
        try:
            with config_path.open('rb') as f:
                logger.debug('Loading MTAP config from "{}"'.format(config_path))
                return _load_config(f)
        except FileNotFoundError:
            pass
    raise ValueError('Failed to load configuration file from')


class Config(MutableMapping[str, Any]):
    """The MTAP configuration dictionary.

    By default configuration is loaded from one of a number of locations in the following priority:

      - A file at the path of the '--config' parameter passed into main methods.
      - A file at the path of the 'MTAP_CONFIG' environment variable
      - $PWD/mtapConfig.yml
      - $HOME/.mtap/mtapConfig.yml'
      - /etc/mtap/mtapConfig.yml


    MTAP components will use a global shared configuration object, by entering the context of a
    config object using "with", all of the MTAP functions called on that thread will make use of
    that config object.

    Examples:

        >>> with mtap.Config() as config:
        >>>     config['key'] = 'value'
        >>>     # other MTAP methods in this
        >>>     # block will use the updated config object.

    """

    _lock = threading.RLock()
    _global_instance = None
    _context = threading.local()
    _context.config = None

    def __new__(cls, *args):
        if cls._global_instance is None:
            with cls._lock:
                if cls._global_instance is None:
                    cls._global_instance = object.__new__(cls)
                    cls._global_instance._config = {}
                    if len(args) == 0:
                        cls._global_instance._load_default_config()
                    else:
                        cls._global_instance._config.update(*args)
        try:
            inst = cls._context.config
        except AttributeError:
            cls._context.config = None
            inst = None
        if inst is not None:
            return inst
        inst = object.__new__(cls)
        inst._config = {}
        inst.update(cls._global_instance)
        return inst

    def __enter__(self) -> 'Config':
        self.enter_context()
        return self

    def enter_context(self):
        with self._lock:
            if hasattr(self._context, "config") and self._context.config is not None:
                raise ValueError("Already in a configuration context.")
            self._context.config = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._context.config = None

    def __reduce__(self):
        return Config, (dict(self), )

    def _load_default_config(self):
        self.update(_load_default_config())

    def update_from_yaml(self, path: Union[Path, str]):
        """Updates the configuration by loading and collapsing all of the structures in a yaml file.

        Parameters:
            path: The path to the yaml file to load.

        """
        if isinstance(path, str):
            path = Path(path)
        with path.open('rb') as f:
            self.update(_load_config(f))

    def __setitem__(self, k: str, v: Any) -> None:
        self._config[k] = v

    def __delitem__(self, v: str) -> None:
        del self._config[v]

    def __getitem__(self, k: str) -> Any:
        return self._config[k]

    def __len__(self) -> int:
        return len(self._config)

    def __iter__(self) -> Iterator[str]:
        return iter(self._config)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Public API and access points for the MTAP Framework."""

from mtap import version
from mtap._config import Config
from mtap._events_service import EventsServer
from mtap.data import Document
from mtap.data import Event
from mtap.data import EventsClient
from mtap.data import GenericLabel
from mtap.data import label
from mtap.data import label_index
from mtap.data import Location
from mtap.processing import descriptions
from mtap.processing import DocumentProcessor
from mtap.processing import EventProcessor
from mtap.processing import processor_parser
from mtap.processing import LocalProcessor
from mtap.processing import Pipeline
from mtap.processing import ProcessorServer
from mtap.processing import RemoteProcessor
from mtap.processing import processor
from mtap.processing import run_processor

__version__ = version.version

__all__ = [s for s in dir()]



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Module for deploying a set of processing services and the events server all at once.

Examples:

    An example configuration

    >>> deploy = Deployment(
    >>>     GlobalSettings(host='0.0.0.0'),
    >>>     EventsDeployment(port=10100, workers=8),
    >>>     SharedProcessorConfig(workers=8, jvm_args=['-Xms32m', '-Xmx8g'], classpath='blah.jar'),
    >>>     ProcessorDeployment(implementation='python',
    >>>                         entry_point='mtap.examples.example_processor',
    >>>                         instances=4,
    >>>                         port=10101,
    >>>                         workers=4),
    >>>     ProcessorDeployment(implementation='java',
    >>>                         entry_point='edu.umn.nlpie.mtap.WordOccurrencesExampleProcessor',
    >>>                         port=10105)
    >>> )
    >>> deploy.run_servers()



"""
import argparse
import logging
import os
import pathlib
import shutil
import subprocess
import sys
import threading
import time
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import grpc

from mtap import utilities, _config

__all__ = [
    'Deployment', 'GlobalSettings', 'SharedProcessorConfig', 'EventsDeployment',
    'ProcessorDeployment', 'main', 'deployment_parser', 'ServiceDeploymentException',
]

logger = logging.getLogger(__name__)

PYTHON_EXE = sys.executable


def _get_java() -> str:
    try:
        return str(pathlib.Path(os.environ['JAVA_HOME']) / 'bin' / 'java')
    except KeyError:
        return 'java'


JAVA_EXE = _get_java()


def _listen(process: subprocess.Popen) -> int:
    for line in process.stdout:
        print(line.decode(), end='', flush=True)
    return process.wait()


class ServiceDeploymentException(Exception):
    """Exception raised in the case of a service failing to launch.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class GlobalSettings:
    """Settings shared by event service and all processors.

    Keyword Args:
        host (Optional[str]): The global host, by default will use "127.0.0.1".
        mtap_config (Optional[str]): The path to an MTAP config file to load for all services.
        log_level (Optional[str]): A python logging level to pass to all services.
        register (Optional[str]): Whether services should register with service discovery.

    Attributes:
        host (Optional[str]): The global host, by default will use "127.0.0.1".
        mtap_config (Optional[str]): The path to an MTAP config file to load for all services.
        log_level (Optional[str]): A python logging level to pass to all services.
        register (Optional[str]): Whether services should register with service discovery.

    """

    def __init__(self, *,
                 host: Optional[str] = None,
                 mtap_config: Optional[str] = None,
                 log_level: Optional[str] = None,
                 register: Optional[bool] = None):
        self.host = host
        self.mtap_config = mtap_config
        self.log_level = log_level
        self.register = register

    @staticmethod
    def from_conf(conf: Optional[Dict]) -> 'GlobalSettings':
        """Creates a global settings object from a configuration dictionary.

        Keyword Args:
            conf (Optional[Dict]): The configuration dictionary.

        Returns:
            GlobalSettings: The global settings object.

        """
        conf = conf or {}
        return GlobalSettings(host=conf.get('host'), mtap_config=conf.get('mtap_config'),
                              log_level=conf.get('log_level'), register=conf.get('register'))


class SharedProcessorConfig:
    """Configuration that is shared between multiple processor services.

    Keyword Args:
        events_addresses (Optional[str]): An optional GRPC-compatible target for the events
            service to be used by all processors.
        workers (Optional[int]): The default number of worker threads which will perform
            processing.
        additional_args (Optional[List[str]]): a list of additional arguments that
            should be appended to every processor.
        jvm_args (Optional[List[str]]): a list of JVM arguments for all java
            processors.
        java_classpath (Optional[str]): A classpath string that will be passed to all java
            processors.
        startup_timeout (Optional[int]): The default startup timeout for processors.
        mp_spawn_method (Optional[str]): A :meth:`multiprocessing.get_context` argument to create
            the multiprocessing context.

    Attributes:
        events_addresses (Optional[List[str]]): An optional GRPC-compatible target for the events
            service to be used by all processors.
        workers (Optional[int]): The default number of worker threads which will perform
            processing.
        additional_args (Optional[List[str]]): a list of additional arguments that
            should be appended to every processor.
        jvm_args (Optional[List[str]]): a list of JVM arguments for all java
            processors.
        java_classpath (Optional[str]): A classpath string that will be passed to all java
            processors.
        startup_timeout (Optional[int]): The default startup timeout for processors.
        mp_spawn_method (Optional[str]): A :meth:`multiprocessing.get_context` argument to create
            the multiprocessing context.

    """

    def __init__(self,
                 events_addresses: Optional[List[str]] = None,
                 workers: Optional[int] = None,
                 additional_args: Optional[List[str]] = None,
                 jvm_args: Optional[List[str]] = None,
                 java_classpath: Optional[str] = None,
                 startup_timeout: Optional[int] = None,
                 mp_spawn_method: Optional[str] = None):
        self.events_addresses = events_addresses
        self.workers = workers
        self.additional_args = additional_args
        self.jvm_args = jvm_args
        self.java_classpath = java_classpath
        self.startup_timeout = startup_timeout or 30
        self.mp_spawn_method = mp_spawn_method

    @staticmethod
    def from_conf(conf: Optional[Dict]) -> 'SharedProcessorConfig':
        """Builds a configuration from a dictionary representation.

        Args:
            conf (Optional[Dict]): The configuration dictionary.

        Returns:
            SharedProcessorConfig object.

        """
        conf = conf or {}
        return SharedProcessorConfig(**conf)


class _ServiceDeployment:
    def __init__(self,
                 workers: Optional[int],
                 register: Optional[bool],
                 mtap_config: Optional[str],
                 log_level: Optional[str]):
        self.workers = workers
        self.register = register
        self.mtap_config = mtap_config
        self.log_level = log_level

    def service_args(self,
                     host: Optional[str] = None,
                     port: Optional[int] = None,
                     register_default: Optional[bool] = None,
                     host_default: Optional[str] = None,
                     workers_default: Optional[int] = None,
                     mtap_config_default: Optional[str] = None,
                     log_level_default: Optional[str] = None):
        call = []

        host = host or host_default
        if host is not None:
            call.extend(['--host', str(host)])

        if port is not None:
            call.extend(['--port', str(port)])

        if self.register or register_default:
            call.append('--register')

        workers = self.workers or workers_default
        if workers is not None:
            call.extend(['--workers', str(workers)])

        mtap_config = self.mtap_config or mtap_config_default
        if mtap_config is not None:
            call.extend(['--mtap-config', mtap_config])

        log_level = self.log_level or log_level_default
        if log_level is not None:
            call.extend(['--log-level', log_level])

        call.append('--write-address')
        return call


class EventsDeployment:
    """Deployment configuration for the events service.

    Keyword Args:
        enabled (bool): Whether an events service should be created.
        addresses (~typing.Optional[~typing.Sequence[str]]): The host address of the events service.
        workers (~typing.Optional[int]): The number of worker threads the events service should use.
        register (~typing.Optional[bool]): Whether to register the events service with discovery.
        mtap_config (~typing.Optional[str]): Path to an mtap configuration file.
        log_level (~typing.Optional[str]): The log level for the events service.

    """

    def __init__(self, *,
                 enabled: bool = True,
                 addresses: Optional[Sequence[str]] = None,
                 workers: Optional[int] = None,
                 register: Optional[bool] = None,
                 mtap_config: Optional[str] = None,
                 log_level: Optional[str] = None):
        self.enabled = enabled
        self.addresses = addresses
        self.service_deployment = _ServiceDeployment(workers, register, mtap_config, log_level)

    def create_calls(self, global_settings: GlobalSettings) -> Iterable[List[str]]:
        for address in self.addresses:
            host = None
            port = None
            if address:
                splits = address.split(':')
                if len(splits) == 2:
                    host, port = splits
                    if host == '':
                        host = None
                else:
                    host = splits[0]
            call = [PYTHON_EXE, '-m', 'mtap', 'events']
            service_args = self.service_deployment.service_args(
                host=host,
                port=port,
                register_default=global_settings.register,
                host_default=global_settings.host,
                mtap_config_default=global_settings.mtap_config,
                log_level_default=global_settings.log_level
            )
            call.extend(service_args)
            yield call

    @staticmethod
    def from_conf(conf: Optional[Dict]) -> 'EventsDeployment':
        """Creates the EventsDeployment configuration option from a configuration dictionary.

        Args:
            conf (Optional[Dict]): The configuration dictionary

        Returns:
            EventsDeployment or None from the configuration dictionary.

        """
        conf = conf or {}

        enabled = conf.get('enabled')
        if enabled is None:
            enabled = False

        address = conf.get('address', None) or conf.get('addresses', None)
        if address is None:
            addresses = []
        elif isinstance(address, str):
            addresses = [address]
        elif isinstance(address, Iterable):
            addresses = list(address)
        else:
            raise ValueError('Unrecognized type of address: ' + type(address))

        return EventsDeployment(enabled=enabled, addresses=addresses,
                                workers=conf.get('workers'), register=conf.get('register'),
                                mtap_config=conf.get('mtap_config'))


class ProcessorDeployment:
    """Deployment configuration for an MTAP processor.

    Used to construct the command for launching the processor. The processor should be a Java Class
    with a main method or a Python module with a main block. It should accept the standard MTAP
    processor deployment arguments and launch an MTAP processor using :func:`mtap.run_processor` or
    the equivalent Java method.

    Args:
        implementation (str): Either "java" or "python".
        entry_point (str): Either the java main class, or the python main module.
        enabled (~typing.Optional[bool]): Whether the processor should be launched as part of
            deployment. Default is `True` if `None`.
        instances (~typing.Optional[int]): The number of instances of the processor to launch.
            Default is `1` if `None`.
        host (~typing.Optional[str]): The listening host for the processor service.
        port (~typing.Optional[int]): The listening port for the processor service.
        workers (~typing.Optional[int]): The number of worker threads per instance.
        register (~typing.Optional[bool]):
            Whether the processor should register with the discovery service specified in the MTAP
            configuration
        mtap_config (~typing.Optional[str]): Path to the MTAP configuration file.
        log_level (~typing.Optional[str]): The log level for the processor.
        identifier (~typing.Optional[str]): An optional identifier override to use for registration.
        pre_args (~typing.Optional[~typing.List[str]]):
            Arguments that occur prior to the MTAP service arguments (like host, port, etc).
        additional_args (~typing.Optional[~typing.List[str]]):
            Arguments that occur after the MTAP service arguments.
        startup_timeout (~typing.Optional[int]): Optional override startup timeout.
        mp_spawn_method (~typing.Optional[str]): A :meth:`multiprocessing.get_context` argument to create
            the multiprocessing context.

    """

    def __init__(self,
                 implementation: str,
                 entry_point: str,
                 *, enabled: Optional[bool] = None,
                 instances: Optional[int] = None,
                 host: Optional[str] = None,
                 port: Optional[int] = None,
                 workers: Optional[int] = None,
                 register: Optional[bool] = None,
                 mtap_config: Optional[str] = None,
                 log_level: Optional[str] = None,
                 identifier: Optional[str] = None,
                 pre_args: Optional[List[str]] = None,
                 additional_args: Optional[List[str]] = None,
                 startup_timeout: Optional[int] = None,
                 mp_spawn_method: Optional[str] = None):
        self.enabled = enabled if enabled is not None else True
        self.implementation = implementation
        self.entry_point = entry_point
        self.instances = instances or 1
        if not isinstance(self.instances, int) or self.instances < 1:
            raise ValueError("Instances must be strictly positive integer.")
        self.identifier = identifier
        self.pre_args = pre_args
        self.additional_args = additional_args
        self.host = host
        self.port = port
        self.service_deployment = _ServiceDeployment(workers, register, mtap_config, log_level)
        self.startup_timeout = startup_timeout
        self.mp_spawn_method = mp_spawn_method

    @staticmethod
    def from_conf(conf: Dict) -> 'ProcessorDeployment':
        """Creates an MTAP processor deployment configuration from a configuration dictionary.

        Args:
            conf (Dict): The configuration dictionary.

        Returns:
            ProcessorDeployment object that can be used to constuct the call for the processor.

        """
        return ProcessorDeployment(**conf)

    def create_calls(self,
                     global_settings: GlobalSettings,
                     shared_config: SharedProcessorConfig) -> Iterable[List[str]]:
        if isinstance(self.port, list):
            ports = self.port
        elif self.port is None:
            ports = [None] * self.instances
        else:
            ports = list(range(self.port, self.port + self.instances))
        for port in ports:
            if self.implementation == 'python':
                call = [PYTHON_EXE, '-m', self.entry_point]
                mp_spawn_method = shared_config.mp_spawn_method
                if self.mp_spawn_method is not None:
                    mp_spawn_method = self.mp_spawn_method
                if mp_spawn_method is not None:
                    call.extend(['--mp-spawn-method', mp_spawn_method])
            elif self.implementation == 'java':
                call = [str(JAVA_EXE)]
                if shared_config.jvm_args is not None:
                    call.extend(shared_config.jvm_args)
                if shared_config.java_classpath is not None:
                    call.extend(['-cp', shared_config.java_classpath])
                call.append(self.entry_point)
            else:
                raise ValueError('Unrecognized implementation: ' + self.implementation)

            if self.pre_args is not None:
                call.extend(self.pre_args)

            service_args = self.service_deployment.service_args(
                host=self.host,
                port=port,
                register_default=global_settings.register,
                host_default=global_settings.host,
                mtap_config_default=global_settings.mtap_config,
                log_level_default=global_settings.log_level,
                workers_default=shared_config.workers
            )
            call.extend(service_args)

            if self.identifier is not None:
                call.extend(['--identifier', self.identifier])

            events_addresses = shared_config.events_addresses
            if events_addresses is not None:
                call.extend(['--events', ','.join(events_addresses)])

            if self.additional_args is not None:
                call.extend(self.additional_args)

            if shared_config.additional_args is not None:
                call.extend(shared_config.additional_args)

            yield call


class Deployment:
    """A automatic deployment configuration which launches a configurable set of MTAP services.

    Args:
         global_settings (~typing.Optional[GlobalSettings]): Settings shared among all services.
         events_deployment (~typing.Optional[EventsDeployment]):
            Deployment settings for the events service.
         shared_processor_config (~typing.Optional[SharedProcessorConfig]):
            Shared configuration settings for all processors.
         processors (vararg ProcessorDeployment): Configurations for individual processors.

    """

    def __init__(self,
                 global_settings: Optional[GlobalSettings] = None,
                 events_deployment: Optional[EventsDeployment] = None,
                 shared_processor_config: Optional[SharedProcessorConfig] = None,
                 *processors: ProcessorDeployment):
        self.global_settings = global_settings
        self.events_deployment = events_deployment
        self.shared_processor_config = shared_processor_config
        self.processors = processors

    @staticmethod
    def load_configuration(conf: Dict) -> 'Deployment':
        """Creates a deployment object from a configuration dictionary.

        Args:
            conf (Dict): The configuration dictionary.

        Returns:
            Deployment object created.

        """
        global_settings = GlobalSettings.from_conf(conf.get('global'))
        events = EventsDeployment.from_conf(conf.get('events_service'))
        shared_processor_config = SharedProcessorConfig.from_conf(conf.get('shared_processor_config'))
        processors_list = conf.get('processors', [])
        processors = [ProcessorDeployment.from_conf(c) for c in processors_list]
        return Deployment(global_settings, events, shared_processor_config, *processors)

    @staticmethod
    def from_yaml_file(conf_path: Union[pathlib.Path, str]) -> 'Deployment':
        """Loads a deployment configuration from a yaml file.

        Args:
            conf_path (str or pathlib.Path): The path to the yaml configuration file.

        Returns:
            Deployment object created from the configuration.

        """
        conf_path = pathlib.Path(conf_path)
        from yaml import load
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader
        with conf_path.open('rb') as f:
            conf = load(f, Loader=Loader)
        return Deployment.load_configuration(conf)

    def run_servers(self):
        """Starts all of the configured services.

        Raises:
            ServiceDeploymentException: If one of the services fails to launch.

        """
        with _config.Config() as c:
            enable_proxy = c.get('grpc.enable_proxy', False)
            processes_listeners = []
            events_addresses = []

            def shutdown(kill=False):
                print("Shutting down all processors")
                for p, listener in processes_listeners:
                    if kill:
                        p.kill()
                    listener.join(timeout=1)

            if self.events_deployment.enabled:
                for call in self.events_deployment.create_calls(self.global_settings):
                    p = subprocess.Popen(call, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                    listener, events_address = _listen_test_connectivity(p, "events", 30, enable_proxy)
                    events_addresses.append(events_address)
                    processes_listeners.append((p, listener))

            if (not self.global_settings.register
                    and not self.events_deployment.service_deployment.register
                    and self.shared_processor_config.events_addresses is None):
                self.shared_processor_config.events_addresses = events_addresses

            for processor_deployment in self.processors:
                if processor_deployment.enabled:
                    for call in processor_deployment.create_calls(self.global_settings,
                                                                  self.shared_processor_config):
                        logger.debug('Launching processor with call: %s', call)
                        p = subprocess.Popen(call, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                        startup_timeout = (processor_deployment.startup_timeout
                                           or self.shared_processor_config.startup_timeout)
                        try:
                            listener, _ = _listen_test_connectivity(p, call, startup_timeout,
                                                                    enable_proxy)
                        except ServiceDeploymentException as e:
                            logger.error(str(e))
                            return shutdown(kill=True)
                        processes_listeners.append((p, listener))

            print('Done deploying all servers.', flush=True)
            try:
                while True:
                    time.sleep(60 * 60 * 24)
            except KeyboardInterrupt:
                return shutdown()


def _listen_test_connectivity(p: subprocess.Popen,
                              name: Any,
                              startup_timeout: int,
                              enable_proxy: bool = False) -> Tuple[threading.Thread, str]:
    listener = threading.Thread(target=_listen, args=(p,))
    listener.start()
    address = None
    for i in range(startup_timeout):
        try:
            address = utilities.read_address(str(p.pid))
            break
        except FileNotFoundError:
            time.sleep(1)
    if address is None:
        raise ServiceDeploymentException('Timed out waiting for {} to launch'.format(name))
    with grpc.insecure_channel(address,
                               options=[('grpc.enable_http_proxy', enable_proxy)]) as channel:
        future = grpc.channel_ready_future(channel)
        try:
            future.result(timeout=startup_timeout)
        except grpc.FutureTimeoutError:
            raise ServiceDeploymentException('Failed to launch: {}'.format(name))
    return listener, address


def main(args: Optional[Sequence[str]] = None,
         conf: Optional[argparse.Namespace] = None):
    if conf is None:
        conf = deployment_parser().parse_args(args)
    if conf.log_level is not None:
        logging.basicConfig(level=getattr(logging, conf.log_level))
    if conf.mode == 'run_servers':
        deployment = Deployment.from_yaml_file(conf.deploy_config)
        deployment.run_servers()
    if conf.mode == 'write_example':
        example = pathlib.Path(__file__).parent / "examples" / "exampleDeploymentConfiguration.yml"
        shutil.copyfile(str(example), "exampleDeploymentConfiguration.yml")
        print('Writing "exampleDeploymentConfiguration.yml" to ' + str(pathlib.Path.cwd()))


def deployment_parser() -> argparse.ArgumentParser:
    """Creates a parser for configuration that can be passed to the deployment main method.

    Returns:
        ~argparse.ArgumentParser: The argument parser object that will create a namespace that can
        be passed to :func:`main`.

    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--log-level', metavar='LEVEL',
                        help="The log level to use for the deployment script.")
    subparsers = parser.add_subparsers(title='mode')

    run_servers = subparsers.add_parser('run_servers')
    run_servers.add_argument('deploy_config', metavar='CONFIG_FILE', type=pathlib.Path,
                             help="A path to the deployment configuration to deploy.")
    run_servers.set_defaults(mode='run_servers')

    write_example = subparsers.add_parser('write_example')
    write_example.set_defaults(mode='write_example')

    return parser


if __name__ == '__main__':
    main()



#  Copyright 2021 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import timedelta
import math
import typing
from typing import TYPE_CHECKING

from mtap.processing import _base

if TYPE_CHECKING:
    import mtap
    from mtap import processing


class TimerStatsAggregator:
    __slots__ = ('_count', '_min', '_max', '_mean', '_sse', '_sum')

    def __init__(self):
        self._count = 0
        self._min = timedelta.max
        self._max = timedelta.min
        self._mean = 0.0
        self._sse = 0.0
        self._sum = timedelta(seconds=0)

    def add_time(self, time):
        if time < self._min:
            self._min = time
        if time > self._max:
            self._max = time

        self._count += 1
        self._sum += time
        time = time.total_seconds()
        delta = time - self._mean
        self._mean += delta / self._count
        delta2 = time - self._mean
        self._sse += delta * delta2

    def finalize(self):
        mean = timedelta(seconds=self._mean)
        variance = self._sse / self._count
        std = math.sqrt(variance)
        std = timedelta(seconds=std)
        return _base.TimerStats(mean=mean, std=std, max=self._max, min=self._min, sum=self._sum)


def add_times(times_map, times):
    for k, v in times.items():
        try:
            agg = times_map[k]
        except KeyError:
            agg = TimerStatsAggregator()
            times_map[k] = agg
        agg.add_time(v)


def create_timer_stats(times_map, prefix=None) -> typing.Dict[str, 'processing.TimerStats']:
    return {identifier: stats.finalize()
            for identifier, stats in times_map.items() if identifier.startswith(prefix)}



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from mtap.data._events import EventsClient, Event, Document


def test_add_document(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    event = Event(event_id='1', client=client)
    document = Document('plaintext',
                        text="“You're no help,” he told the lime. "
                             "This was unfair. It was only a lime; "
                             "there was nothing special about it at all. "
                             "It was doing the best it could.")
    event.add_document(document)
    assert event.documents['plaintext'] == document
    client.add_document.assert_called_once_with('1', 'plaintext',
                                                "“You're no help,” he told the lime. "
                                                "This was unfair. It was only a lime; "
                                                "there was nothing special about it at all. "
                                                "It was doing the best it could.")


def test_add_document_no_client():
    event = Event(event_id='1')
    document = Document('plaintext',
                        text="“You're no help,” he told the lime. "
                             "This was unfair. It was only a lime; "
                             "there was nothing special about it at all. "
                             "It was doing the best it could.")
    event.add_document(document)
    assert event.documents['plaintext'] == document


def test_create_document(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    event = Event(event_id='1', client=client)
    document = event.create_document('plaintext', "“You're no help,” he told the lime. "
                                                  "This was unfair. It was only a lime; "
                                                  "there was nothing special about it at all. "
                                                  "It was doing the best it could.")
    assert event.documents['plaintext'] == document
    client.add_document.assert_called_once_with('1', 'plaintext',
                                                "“You're no help,” he told the lime. "
                                                "This was unfair. It was only a lime; "
                                                "there was nothing special about it at all. "
                                                "It was doing the best it could.")


def test_create_document_no_client():
    event = Event(event_id='1')
    document = event.create_document('plaintext', "“You're no help,” he told the lime. "
                                                  "This was unfair. It was only a lime; "
                                                  "there was nothing special about it at all. "
                                                  "It was doing the best it could.")
    assert event.documents['plaintext'] == document


def test_get_document(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext']
    event = Event(event_id='1', client=client)
    document = event.documents['plaintext']
    assert document.document_name == 'plaintext'


def test_get_document_missing(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext']
    event = Event(event_id='1', client=client)
    with pytest.raises(KeyError):
        _ = event.documents['foo']


def test_get_document_missing_no_client(mocker):
    event = Event(event_id='1')
    with pytest.raises(KeyError):
        _ = event.documents['foo']


def test_iter_documents(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext', 'two', 'three']
    event = Event(event_id='1', client=client)
    documents = {document for document in event.documents}
    assert documents == {'plaintext', 'two', 'three'}


def test_contains_documents(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext', 'two', 'three']
    event = Event(event_id='1', client=client)
    assert 'plaintext' in event.documents
    assert 'two' in event.documents
    assert 'four' not in event.documents


def test_documents_len(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext', 'two', 'three']
    event = Event(event_id='1', client=client)
    assert len(event.documents) == 3


def test_created_indices_no_documents():
    event = Event(event_id='1')
    assert event.created_indices == {}



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import concurrent.futures.thread as thread
import logging
import threading
import time
import traceback
import typing
from typing import Any, Dict, Tuple, Sequence, Optional, Union

import google.protobuf.empty_pb2 as empty_pb2
import grpc
import grpc_health.v1.health as health
import grpc_health.v1.health_pb2_grpc as health_pb2_grpc

from mtap import _config
from mtap import _structs
from mtap import data
from mtap import utilities
from mtap.api.v1 import processing_pb2, processing_pb2_grpc
from mtap.processing import _base, _runners, _timing

if typing.TYPE_CHECKING:
    import mtap

logger = logging.getLogger(__name__)
_ONE_DAY_IN_SECONDS = 60 * 60 * 24


def run_processor(proc: 'mtap.EventProcessor',
                  *,
                  mp: bool = False,
                  options: Optional[argparse.Namespace] = None,
                  args: Optional[Sequence[str]] = None,
                  mp_context=None):
    """Runs the processor as a GRPC service, blocking until an interrupt signal is received.

    Args:
        proc (EventProcessor): The processor to host.
        mp (bool): If true, will create instances of ``proc`` on multiple forked processes to
            process events. This is useful if the processor is computationally intensive and would
            run into Python GIL issues on a single process.
        options (~typing.Optional[~argparse.Namespace]): The parsed arguments from the parser
            returned by :func:`processor_parser`.
        args (~typing.Optional[~typing.Sequence[str]]): Arguments to parse server settings from if
            ``namespace`` was not supplied.
        mp_context: A multiprocessing context that gets passed to the process pool executor in the
            case of mp = True.

    Examples:
        Will automatically parse arguments:

        >>> run_processor(MyProcessor())

        Manual arguments:

        >>> run_processor(MyProcessor(), args=['-p', '8080'])


    """
    from mtap import EventProcessor
    if not isinstance(proc, EventProcessor):
        raise ValueError("Processor must be instance of EventProcessor class.")

    if options is None:
        processors_parser = argparse.ArgumentParser(parents=[processor_parser()])
        processors_parser.add_help = True
        options = processors_parser.parse_args(args)

    if options.log_level:
        logging.basicConfig(level=getattr(logging, options.log_level))

    events_addresses = []
    if options.events_addresses is not None:
        events_addresses.extend(options.events_addresses.split(','))

    with _config.Config() as c:
        if options.mtap_config is not None:
            c.update_from_yaml(options.mtap_config)
        # instantiate runner
        processor_id = options.identifier or proc.metadata['name']

        enable_http_proxy = options.grpc_enable_http_proxy
        if enable_http_proxy is not None:
            c['grpc.enable_proxy'] = enable_http_proxy
        if mp:
            runner = MpProcessorRunner(proc=proc,
                                       workers=options.workers,
                                       events_address=events_addresses,
                                       identifier=processor_id,
                                       mp_context=mp_context)
        else:
            client = data.EventsClient(address=events_addresses)
            runner = _runners.ProcessorRunner(proc, client=client,
                                              identifier=processor_id,
                                              params=None)
        server = ProcessorServer(runner=runner,
                                 host=options.host,
                                 port=options.port,
                                 register=options.register,
                                 workers=options.workers,
                                 write_address=options.write_address)
        server.start()
        try:
            while True:
                time.sleep(_ONE_DAY_IN_SECONDS)
        except KeyboardInterrupt:
            server.stop()


_mp_processor = ...  # EventProcessor
_mp_client = ...  # EventClient


def _mp_initialize(proc: 'mtap.EventProcessor', events_address, config):
    global _mp_processor
    global _mp_client
    _config.Config(config)
    _mp_processor = proc
    _mp_client = data.EventsClient(address=events_address)


def _mp_call_process(event_id, event_service_instance_id, params):
    global _mp_processor
    global _mp_client
    with _base.Processor.enter_context() as c, \
            data.Event(event_id=event_id, event_service_instance_id=event_service_instance_id,
                       client=_mp_client) as event:
        with _base.Processor.started_stopwatch('process_method'):
            result = _mp_processor.process(event, params)
        return result, c.times, event.created_indices


class MpProcessorRunner:
    __slots__ = ('pool', 'metadata', 'processor_id', 'component_id')

    def __init__(self,
                 proc: 'mtap.EventProcessor',
                 identifier: str,
                 workers: 'Optional[int]' = 8,
                 events_address: 'Optional[Union[str, Sequence[str]]]' = None,
                 mp_context=None):
        if mp_context is None:
            import multiprocessing as mp
            mp_context = mp
        config = _config.Config()
        self.pool = mp_context.Pool(workers,
                                    initializer=_mp_initialize,
                                    initargs=(proc, events_address, dict(config)))
        self.metadata = proc.metadata
        self.processor_id = identifier
        self.component_id = identifier

    def call_process(self,
                     event_id: str,
                     event_service_instance_id: str,
                     params: Optional[Dict[str, Any]]) -> Tuple[Dict, Dict, Dict]:
        p = dict()
        if params is not None:
            p.update(params)

        return self.pool.apply(_mp_call_process, args=(event_id, event_service_instance_id, p))

    def close(self):
        self.pool.terminate()
        self.pool.join()


def processor_parser() -> argparse.ArgumentParser:
    """An :class:`~argparse.ArgumentParser` that can be used to parse the settings for
    :func:`run_processor`.

    Returns:
        ~argparse.ArgumentParser: A parser containing server settings.

    Examples:
        Using this as a parent parser:

        >>> parser = ArgumentParser(parents=[processor_parser()])
        >>> parser.add_argument('--my-arg-1')
        >>> parser.add_argument('--my-arg-2')
        >>> args = parser.parse_args()
        >>> processor = MyProcessor(args.my_arg_1, args.my_arg_2)
        >>> run_processor(processor, args)

    """
    processors_parser = argparse.ArgumentParser(add_help=False)
    processors_parser.add_argument('--host', '--address', '-a', default="127.0.0.1", metavar="HOST",
                                   help='the address to serve the service on')
    processors_parser.add_argument('--port', '-p', type=int, default=0, metavar="PORT",
                                   help='the port to serve the service on')
    processors_parser.add_argument('--workers', '-w', type=int, default=10,
                                   help='number of worker threads to handle requests')
    processors_parser.add_argument('--register', '-r', action='store_true',
                                   help='whether to register the service with the configured '
                                        'service discovery')
    processors_parser.add_argument("--mtap-config", default=None,
                                   help="path to MTAP config file")
    processors_parser.add_argument('--events-addresses', '--events-address', '--events', '-e',
                                   default=None,
                                   help='address of the events service to use, '
                                        'omit to use discovery')
    processors_parser.add_argument('--identifier', '-i',
                                   help="Optional argument if you want the processor to register "
                                        "under a different identifier than its name.")
    processors_parser.add_argument('--write-address', action='store_true',
                                   help='If set, will write the server address ')
    processors_parser.add_argument('--log-level', type=str, default='INFO',
                                   help="Sets the python log level.")
    processors_parser.add_argument('--grpc-enable-http-proxy', action='store_true',
                                   help="If set, will enable usage of http_proxy by grpc.")
    processors_parser.add_argument('--mp-spawn-method',
                                   choices=['spawn', 'fork', 'forkserver', None],
                                   help="A multiprocessing spawn method to use.")
    return processors_parser


def _label_index_meta_to_proto(d, message):
    message.name = d['name'] or ''
    message.name_from_parameter = d['name_from_parameter'] or ''
    message.optional = d.get('optional', False)
    message.description = d['description'] or ''
    for property_meta in d['properties']:
        p = message.properties.add()
        p.name = property_meta['name'] or ''
        p.description = property_meta['description'] or ''
        p.data_type = property_meta['data_type'] or ''
        p.nullable = property_meta['nullable']


class _ProcessorServicer(processing_pb2_grpc.ProcessorServicer):
    def __init__(self,
                 config: 'mtap.Config',
                 address: str,
                 runner: '_base.ProcessingComponent',
                 health_servicer: health.HealthServicer,
                 register: bool = False):
        self.config = config
        self.address = address
        self._runner = runner
        self.register = register
        self.health_servicer = health_servicer

        self._times_map = {}
        self._deregister = None
        self.processed = 0
        self.failure_count = 0

    def start(self, port: int):

        self.health_servicer.set(self._runner.processor_id, 'SERVING')

        if self.register:
            from mtap._discovery import Discovery
            service_registration = Discovery(config=self.config)
            self._deregister = service_registration.register_processor_service(
                self.address,
                port,
                self._runner.component_id,
                'v1'
            )

    def shutdown(self):
        self.health_servicer.set(self._runner.processor_id, 'NOT_SERVING')
        if self._deregister is not None:
            self._deregister()
        self._runner.close()

    def Process(self, request, context=None):
        event_id = request.event_id
        event_service_instance_id = request.event_service_instance_id
        logger.debug(
            'processor_id: %s received process request on event_id: %s with event_service_instance_id: %s',
            self._runner.processor_id, event_id, event_service_instance_id)
        params = {}
        _structs.copy_struct_to_dict(request.params, params)
        try:
            response = processing_pb2.ProcessResponse()
            result, times, added_indices = self._runner.call_process(
                event_id,
                event_service_instance_id,
                params
            )
            if result is not None:
                _structs.copy_dict_to_struct(result, response.result, [])

            _timing.add_times(self._times_map, times)
            for k, l in times.items():
                response.timing_info[k].FromTimedelta(l)
            for document_name, l in added_indices.items():
                for index_name in l:
                    created_index = response.created_indices.add()
                    created_index.document_name = document_name
                    created_index.index_name = index_name
            return response
        except Exception as e:
            logger.error(str(e))
            logger.error(traceback.format_exc())
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details(str(e))
            return empty_pb2.Empty()

    def GetStats(self, request, context):
        r = processing_pb2.GetStatsResponse(processed=self.processed,
                                            failures=self.failure_count)
        for k, v in _timing.create_timer_stats(self._times_map, self._runner.component_id).items():
            ts = r.timing_stats[k]
            ts.mean.FromTimedelta(v.mean)
            ts.std.FromTimedelta(v.std)
            ts.max.FromTimedelta(v.max)
            ts.min.FromTimedelta(v.min)
            ts.sum.FromTimedelta(v.sum)
        return r

    def GetInfo(self, request, context):
        response = processing_pb2.GetInfoResponse()
        _structs.copy_dict_to_struct(self._runner.metadata, response.metadata)
        return response


class ProcessorServer:
    """Host a MTAP processor as a service.

    Args:
        proc (EventProcessor): The event processor to host.
        host (str): The address / hostname / IP to host the server on.
        port (int): The port to host the server on, or 0 to use a random port.

    Keyword Args:
        register (~typing.Optional[bool]): Whether to register the processor with service discovery.
        events_address (~typing.Optional[str]):
            The address of the events server, or omitted / None if the events service should be
            discovered.
        processor_id (~typing.Optional[str]):
            The identifier to register the processor under, if omitted the processor name will be
            used.
        workers (~typing.Optional[int]):
            The number of workers that should handle requests. Defaults to 10.
        params (~typing.Optional[~typing.Mapping[str, ~typing.Any]):
            A set of default parameters that will be passed to the processor every time it runs.
        grpc_enable_http_proxy (bool):
            Enables or disables the grpc channel to the event service using http_proxy or
            https_proxy environment variables.
    """

    def __init__(self,
                 runner: 'mtap.processing.ProcessingComponent',
                 host: str,
                 port: int = 0,
                 *,
                 register: bool = False,
                 workers: Optional[int] = None,
                 write_address: bool = False,
                 config: 'Optional[mtap.Config]' = None):
        self.host = host
        self._port = port
        self.processor_id = runner.processor_id
        self.write_address = write_address

        if config is None:
            config = _config.Config()

        self._health_servicer = health.HealthServicer()
        self._health_servicer.set('', 'SERVING')
        self._servicer = _ProcessorServicer(
            config=config,
            address=host,
            runner=runner,
            health_servicer=self._health_servicer,
            register=register
        )
        workers = workers or 10
        thread_pool = thread.ThreadPoolExecutor(max_workers=workers)
        options = config.get("grpc.processor_server_options", {})
        self._server = grpc.server(thread_pool, options=list(options.items()))
        health_pb2_grpc.add_HealthServicer_to_server(self._health_servicer, self._server)
        processing_pb2_grpc.add_ProcessorServicer_to_server(self._servicer, self._server)
        self._port = self._server.add_insecure_port("{}:{}".format(self.host, self.port))
        self._stopped_event = threading.Event()
        self._address_file = None

    @property
    def port(self) -> int:
        """int: Port the hosted server is bound to.
        """
        return self._port

    def start(self):
        """Starts the service.
        """
        self._server.start()
        self._servicer.start(self.port)
        if self.write_address:
            self._address_file = utilities.write_address_file('{}:{}'.format(self.host, self.port))
        logger.info('Started processor server with id: "%s"  on address: "%s:%d"',
                    self.processor_id, self.host, self.port)

    def stop(self, *, grace: Optional[float] = None) -> threading.Event:
        """De-registers (if registered with service discovery) the service and immediately stops
        accepting requests, completely stopping the service after a specified `grace` time.

        During the grace period the server will continue to process existing requests, but it will
        not accept any new requests. This function is idempotent, multiple calls will shutdown
        the server after the soonest grace to expire, calling the shutdown event for all calls to
        this function.

        Keyword Args:
            grace (~typing.Optional[float]):
                The grace period that the server should continue processing requests for shutdown.

        Returns:
            threading.Event: A shutdown event for the server.
        """
        print('Shutting down processor server with id: "{}"  on address: "{}:{}"'.format(
            self.processor_id, self.host, self.port))
        if self._address_file is not None:
            self._address_file.unlink()
        self._servicer.shutdown()
        return self._server.stop(grace=grace)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Internal processors and pipelines functionality."""
import contextlib
import threading
from abc import ABCMeta, abstractmethod, ABC
from collections import Callable
from datetime import timedelta, datetime
from typing import List, ContextManager, Any, Dict, NamedTuple, Optional, Mapping, Generator, \
    Tuple, TYPE_CHECKING

from mtap import data

if TYPE_CHECKING:
    import mtap
    import mtap.processing as processing


class ProcessorContext:
    __slots__ = ('times',)

    def __init__(self):
        self.times = {}

    def add_time(self, key, duration):
        try:
            self.times[key] += duration
        except KeyError:
            self.times[key] = duration


class Stopwatch(ContextManager, metaclass=ABCMeta):
    """A class for timing runtime of components and returning the total runtime with the
    processor's results.

    Attributes:
        duration (~datetime.timedelta): The amount of time elapsed for this timer.

    Examples:
        >>> # in an EventProcessor or DocumentProcessor process method call
        >>> with self.started_stopwatch('key'):
        >>>     timed_routine()


        >>> # in an EventProcessor or DocumentProcessor process method call
        >>> with self.unstarted_stopwatch('key') as stopwatch:
        >>>     for _ in range(10):
        >>>         # work you don't want timed
        >>>         ...
        >>>         stopwatch.start()
        >>>         # work you want timed
        >>>         ...
        >>>         stopwatch.stop()
    """
    __slots__ = ('_key', '_context', '_running', 'duration', '_start')

    def __init__(self, context: Optional = None, key: Optional[str] = None):
        self._key = key
        self._context = context
        self._running = False
        self.duration = timedelta()
        self._start = None

    def start(self):
        """Starts the timer.
        """
        if not self._running:
            self._running = True
            self._start = datetime.now()

    def stop(self):
        """Stops / pauses the timer
        """
        if self._running:
            self._running = False
            self.duration += datetime.now() - self._start

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._running:
            self.stop()
        try:
            self._context.add_time(self._key, self.duration)
        except AttributeError:  # If context is None
            pass


processor_local = threading.local()


class Processor:
    """Mixin used by all processor abstract base classes that provides the ability to update
    serving status and use timers.
    """
    __slots__ = ('_health_callback',)

    metadata = {}

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._health_callback = None
        return instance

    def update_serving_status(self, status: str):
        """Updates the serving status of the processor for health checking.

        Args:
            status (str): One of "SERVING", "NOT_SERVING", "UNKNOWN".
        """
        if self._health_callback is not None:
            self._health_callback(status)

    @property
    def custom_label_adapters(self) -> Mapping[str, data.ProtoLabelAdapter]:
        """Used to provide non-standard proto label adapters for specific index names.

        Returns:

        """
        return {}

    @staticmethod
    def current_context() -> Optional['processing.ProcessorContext']:
        return getattr(processor_local, 'context', None)

    @staticmethod
    def started_stopwatch(key: str) -> Stopwatch:
        """An object that can be used to time aspects of processing. The stopwatch will be started
        at creation.

        Args:
            key (str): The key to store the time under.

        Returns:
            Stopwatch: An object that is used to do the timing.

        Examples:
            >>> # In a process method
            >>> with self.started_stopwatch('key'):
            >>>     # do work
            >>>     ...

        """
        stopwatch = Stopwatch(getattr(processor_local, 'context', None), key)
        stopwatch.start()
        return stopwatch

    @staticmethod
    def unstarted_stopwatch(key: str) -> 'processing.Stopwatch':
        """An object that can be used to time aspects of processing. The stopwatch will be stopped
        at creation.

        Args:
            key (str): The key to store the time under.

        Returns:
            Stopwatch: An object that is used to do the timing.

        Examples:
            >>> # In a process method
            >>> with self.unstarted_stopwatch('key') as stopwatch:
            >>>     for _ in range(10):
            >>>         # work you don't want timed
            >>>         ...
            >>>         stopwatch.start()
            >>>         # work you do want timed
            >>>         ...
            >>>         stopwatch.stop()
        """
        return Stopwatch(getattr(processor_local, 'context', None), key)

    @staticmethod
    @contextlib.contextmanager
    def enter_context() -> ContextManager[ProcessorContext]:
        # Used by the MTAP framework to enter a processing context where things like timing
        # results are stored. Users should not need to call this in normal usage.
        try:
            old_context = processor_local.context
        except AttributeError:
            old_context = None
        try:
            processor_local.context = ProcessorContext()
            yield processor_local.context
        finally:
            del processor_local.context
            if old_context is not None:
                processor_local.context = old_context


class EventProcessor(Processor):
    """Abstract base class for an event processor.

    Examples:
        >>> class ExampleProcessor(EventProcessor):
        ...     def process(self, event, params):
        ...          # do work on the event
        ...          ...
    """
    __slots__ = ()

    @abstractmethod
    def process(self, event: 'mtap.Event', params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Performs processing on an event, implemented by the subclass.

        Parameters:
            event (Event): The event object to be processed.
            params (~typing.Dict[str, ~typing.Any]):
                Processing parameters. A dictionary of strings mapped to json-serializable values.

        Returns:
            ~typing.Optional[~typing.Dict[str, Any]]:
                An arbitrary dictionary of strings mapped to json-serializable values which will be
                returned to the caller, even remotely.
        """
        ...

    def default_adapters(self) -> Optional[Mapping[str, data.ProtoLabelAdapter]]:
        """Can be overridden to return a mapping from label index names to adapters that will then
        be used in any documents or events provided to this processor's process methods.

        Returns:
            Mapping[str, ProtoLabelAdapter]: The mapping (dict-like) of label index names to
            ProtoLabelAdapters to be used for those indices.

        """
        pass

    def close(self):
        """Can be overridden for cleaning up anything that needs to be cleaned up. Will be called
        by the framework after it's done with the processor."""
        pass


class DocumentProcessor(EventProcessor):
    """Abstract base class for a document processor.

    Examples:
        >>> class ExampleProcessor(mtap.DocumentProcessor):
        ...     def process(self, document, params):
        ...         # do processing on document
        ...         ...
    
    
        >>> class ExampleProcessor(mtap.DocumentProcessor):
        ...     def process(self, document, params):
        ...          with self.started_stopwatch('key'):
        ...               # use stopwatch on something
        ...               ...

    """
    __slots__ = ()

    @abstractmethod
    def process_document(self,
                         document: 'mtap.Document',
                         params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Performs processing of a document on an event, implemented by the subclass.

        Args:
            document (~mtap.Document): The document object to be processed.
            params (~typing.Dict[str, ~typing.Any]):
                Processing parameters. A dictionary of strings mapped to json-serializable values.

        Returns:
            ~typing.Dict[str, ~typing.Any]:
                An arbitrary dictionary of strings mapped to json-serializable values that will be
                returned to the caller of the processor.
        """
        ...

    def close(self):
        """Can be overridden for cleaning up anything that needs to be cleaned up. Will be called
        by the framework after it's done with the processor."""
        pass

    def process(self, event: 'mtap.Event', params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        document = event.documents[params['document_name']]
        return self.process_document(document, params)


class ProcessingError(Exception):
    """Exception for when processing results in a failure."""
    pass


ProcessingResult = NamedTuple('ProcessingResult',
                              [('identifier', str),
                               ('result_dict', Dict),
                               ('timing_info', Dict),
                               ('created_indices', Dict[str, List[str]])])

ProcessingResult.__doc__ = """The result of processing one document or event."""
ProcessingResult.identifier.__doc__ = """The id of the processor with respect to the pipeline."""
ProcessingResult.result_dict.__doc__ = """The json object returned by the processor as its results."""
ProcessingResult.timing_info.__doc__ = """A dictionary of the times taken processing this 
document"""
ProcessingResult.created_indices.__doc__ = """Any indices that have been added to documents by this 
processor."""


class PipelineResult(NamedTuple('PipelineResult',
                                [('component_results', List[ProcessingResult]),
                                 ('elapsed_time', timedelta)])):
    """The result of processing an event or document in a pipeline.

    Attributes:
        component_results (List[ProcessingResult]): The processing results for each individual
            component
        elapsed_time (timedelta): The elapsed time for the entire pipeline.

    Args:
        component_results (List[ProcessingResult]): The processing results for each individual
            component
        elapsed_time (timedelta): The elapsed time for the entire pipeline.

    """
    __slots__ = ()

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def component_result(self, identifier: str) -> ProcessingResult:
        """Returns the component result for a specific identifier.

        Args:
            identifier: The processor's identifier in the pipeline.

        Returns:
            ProcessingResult: The result for the specified processor.

        """
        try:
            return next(filter(lambda x: x.identifier == identifier, self.component_results))
        except StopIteration:
            raise KeyError('No result for identifier: ' + identifier)


TimerStats = NamedTuple('TimerStats',
                        [('mean', timedelta),
                         ('std', timedelta),
                         ('min', timedelta),
                         ('max', timedelta),
                         ('sum', timedelta)])
TimerStats.__doc__ = """Statistics about a specific keyed measured duration recorded by a 
:obj:`~mtap.processing.base.Stopwatch`."""
TimerStats.mean.__doc__ = """The sample mean of all measured durations."""
TimerStats.std.__doc__ = """The sample standard deviation of all measured 
durations."""
TimerStats.max.__doc__ = """The minimum of all measured durations."""
TimerStats.min.__doc__ = """The maximum of all measured durations."""
TimerStats.sum.__doc__ = """The sum of all measured durations."""


class AggregateTimingInfo(NamedTuple('AggregateTimingInfo',
                                     [('identifier', str),
                                      ('timing_info', Dict[str, 'processing.TimerStats'])])):
    """Collection of all of the timing info for a specific processor.

    Attributes:
        identifier (str): The ID of the processor with respect to the pipeline.
        timing_info (dict[str, TimerStats]):
            A map from all of the timer keys for the processor to the aggregated duration statistics.
    """
    __slots__ = ()

    def print_times(self):
        """Prints the aggregate timing info for all processing components using ``print``.
        """
        print(self.identifier)
        print("-------------------------------------")
        for key, stats in self.timing_info.items():
            print("  [{}]\n"
                  "    mean: {}\n"
                  "    std: {}\n"
                  "    min: {}\n"
                  "    max: {}\n"
                  "    sum: {}".format(key, stats.mean, stats.std, stats.min,
                                       stats.max, stats.sum))
        print("")

    @staticmethod
    def csv_header() -> str:
        """Returns the header for CSV formatted timing data.

        Returns:
             str
        """
        return 'key,mean,std,min,max,sum\n'

    def timing_csv(self) -> Generator[str, None, None]:
        """Returns the timing data formatted as a string, generating each

        Returns:
            Generator[str]
        """
        for key, stats in self.timing_info.items():
            yield '{}:{},{},{},{},{},{}\n'.format(self.identifier, key, stats.mean, stats.std,
                                                  stats.min, stats.max, stats.sum)


class ProcessingComponent(ABC):
    __slots__ = ()

    metadata = {}

    @abstractmethod
    def call_process(self, event_id: str, event_instance_id: str,
                     params: Optional[Dict[str, Any]]) -> Tuple[Dict, Dict, Dict]:
        """Calls a processor.

        Parameters
        ----------
        event_id: str
            The event to process.
        event_instance_id: str
            The service instance the event is stored on.
        params: Dict
            The processor parameters.

        Returns
        -------
        tuple of dict, dict, dict
            A tuple of the processing result dictionary, the processor times dictionary, and the
            created indices dictionary.

        """
        ...

    def close(self):
        ...


class ComponentDescriptor(ABC):
    """A configuration which describes either a local or remote pipeline component and what the
    pipeline needs to do to call the component.
    """
    __slots__ = ()

    @abstractmethod
    def create_pipeline_component(
            self,
            component_ids: Dict[str, int],
            client: 'Callable[[], mtap.EventsClient]'
    ) -> 'processing.ProcessingComponent':
        pass



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import timedelta

from mtap.processing import Processor


def test_stopwatch_no_fail_outside_context():
    blah = False
    with Processor.started_stopwatch('foo'):
        blah = True
    assert blah


def test_preserves_times():
    with Processor.enter_context() as context:
        context.add_time("foo", timedelta(seconds=2))
        context.add_time("foo", timedelta(seconds=2))
        assert context.times["foo"] == timedelta(seconds=4)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A framework for developing NLP pipeline components."""

import os

import pkg_resources
import sys
from distutils.command.clean import clean as _clean
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.test import test as _test
from pathlib import Path

GRPC_VERSION = '1.47.0'


def generate_proto(source, require=True):
    """Invokes the grpc_tools protobuf compiler to generate _pb2.py and _pb2_grpc.py files from
    the given .proto file.  Does nothing if the output already exists and is newer than
    the input."""

    if not require and not os.path.exists(source):
        return

    output = source.replace(".proto", "_pb2.py").replace("proto/", "")

    if (not os.path.exists(output) or
            (os.path.exists(source) and
             os.path.getmtime(source) > os.path.getmtime(output))):
        print("Generating %s..." % output)

        if not os.path.exists(source):
            sys.stderr.write("Can't find required file: %s\n" % source)
            sys.exit(-1)

        try:
            import grpc_tools.protoc
        except ImportError:
            sys.stderr.write("Can't find grpcio-tools, install using pip")
            sys.exit(-1)

        proto_include = pkg_resources.resource_filename('grpc_tools', '_proto')
        grpc_tools.protoc.main(["-I.",
                                "-I{}".format(proto_include),
                                "-Ithird_party/api-common-protos-0.1.0",
                                "-Iproto",
                                "--python_out=python",
                                "--grpc_python_out=python",
                                source])


class clean(_clean):
    def run(self):
        # Delete generated files in the code tree.
        for (dirpath, dirnames, filenames) in os.walk("python"):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if filepath.endswith("_pb2.py") or filepath.endswith(".pyc") or \
                        filepath.endswith("_pb2_grpc.py"):
                    os.remove(filepath)
        super().run()


class build_py(_build_py):
    def run(self):
        generate_proto('proto/mtap/api/v1/events.proto')
        generate_proto('proto/mtap/api/v1/processing.proto')
        super().run()


class test(_test):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        _test.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with (Path(__file__).parent / 'README.md').open(encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mtap',
    use_scm_version={
        "fallback_version": "development0",
        "write_to": "python/mtap/version.py"
    },
    description='A framework for distributed text analysis using gRPC and microservices-based '
                'architecture.',
    url='https://nlpie.github.io/mtap',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='~=3.5',
    author='University of Minnesota NLP/IE Group',
    author_email='nlp-ie@umn.edu',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Linguistic'
    ],
    keywords='nlp grpc microservices',
    package_dir={'': 'python'},
    packages=find_packages(where='python', exclude=['tests']),
    package_data={
        'mtap': [
            'defaultConfig.yml',
            'examples/exampleDeploymentConfiguration.yml',
            'examples/examplePipelineConfiguration.yml',
        ]
    },
    install_requires=[
        'grpcio~=' + GRPC_VERSION,
        'grpcio-health-checking~=' + GRPC_VERSION,
        'googleapis-common-protos',
        'pyyaml',
        'python-consul',
        'tqdm'
    ],
    setup_requires=[
        'pytest-runner',
        'grpcio-tools~=' + GRPC_VERSION,
        'setuptools_scm',
    ],
    tests_require=[
        'pytest',
        'pytest-mock',
        'grpcio-testing~=' + GRPC_VERSION,
        'requests'
    ],
    extras_require={
        'grpc_tools': ['grpcio-tools~=' + GRPC_VERSION],
        'tests': ['pytest-runner', 'pytest', 'grpcio-testing~=' + GRPC_VERSION, 'requests', 'pytest-mock'],
        'docs': ['sphinx', 'sphinx_rtd_theme']
    },
    cmdclass={
        'clean': clean,
        'build_py': build_py,
        'test': test
    }
)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import typing

import mtap


__all__ = [
    'copy_document', 'CopyDocument', 'main'
]


def copy_document(event: mtap.Event,
                  source_document_name: str,
                  target_document_name: str,
                  index_names: typing.Sequence[str] = ...):
    """Copies one document to another on the same event.

    Parameters
    ----------
    event: Event
        The event.
    source_document_name: str
        The source document name.
    target_document_name: str
        The target document name.
    index_names: Sequence[str]
        If specified will only copy the specified label indices, by default all indices will be
        copied.
    """
    source_document = event.documents[source_document_name]
    target_document = mtap.Document(target_document_name, text=source_document.text)
    event.add_document(target_document)
    if index_names is ...:
        index_names = list(source_document.labels)
    for index_name in index_names:
        index = source_document.labels[index_name]
        target_document.add_labels(index_name, index, distinct=index.distinct)


@mtap.processor('mtap-copy-processor')
class CopyDocument(mtap.EventProcessor):
    """Copies one document to another.

    Parameters
    ----------
    source_document_name: str
        The source document name.
    target_document_name: str
        The target document name.
    index_names: Sequence[str]
        If specified will only copy the specified label indices, by default all indices will be
        copied.
    """

    def __init__(self,
                 source_document_name: str,
                 target_document_name: str,
                 index_names: typing.Sequence[str] = ...):
        self.source_document_name = source_document_name
        self.target_document_name = target_document_name
        self.index_names = index_names

    def process(self, event: mtap.Event, params: typing.Dict[str, typing.Any]):
        copy_document(event, self.source_document_name, self.target_document_name, self.index_names)


def main(args=None):
    parser = argparse.ArgumentParser(
        description='Deploys a processor that copies one document to another.',
        parents=[mtap.processor_parser()]
    )
    parser.add_argument('--index-names', nargs='*', metavar='INDEX_NAME', default=...,
                        help='')
    parser.add_argument('source_document_name', metavar='SOURCE_NAME',
                        help='Name of source document.')
    parser.add_argument('target_document_name', metavar='TARGET_NAME',
                        help='Name of target document.')
    ns = parser.parse_args(args)
    p = CopyDocument(ns.source_document_name, ns.target_document_name, ns.index_names)
    mtap.run_processor(p, options=ns)


if __name__ == '__main__':
    main()



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from mtap import GenericLabel, Document

document = Document('plaintext', text='foo bar')


def test_get_repr():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)

    rep = repr(label)
    assert rep.startswith("GenericLabel(0, 20, ")
    assert 'a="x"' in rep or "a='x'" in rep
    assert 'y=20' in rep
    assert 'z=20.0' in rep


def test_get_repr_ref():
    label = GenericLabel(0, 20, document=document, a=GenericLabel(0, 20, x="a"))

    rep = repr(label)
    assert rep == 'GenericLabel(0, 20, a=GenericLabel(0, 20, x=\'a\'))'


def test_get_repr_infinite_recursion():
    label = GenericLabel(0, 20, document=document)
    b = GenericLabel(0, 20, document=document, a=label)
    label.b = b

    rep = repr(label)
    assert rep == 'GenericLabel(0, 20, b=GenericLabel(0, 20, a=GenericLabel(...)))'


def test_get_attr():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)

    assert label.a == 'x'
    assert label.y == 20
    assert label.z == 20.0
    assert label.fields == {
        'a': 'x',
        'y': 20,
        'z': 20.0
    }


def test_loop_in_map():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)
    with pytest.raises(ValueError):
        label.bar = {
            'label': label
        }


def test_setattr():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)
    label.bar = 'baz'
    assert 'bar' in label.fields
    assert label.fields['bar'] == 'baz'


def test_eq_not_generic_label():
    assert GenericLabel(0, 20, document=document, a="x", y=20, z=20.0) != 0


def test_eq():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)
    label2 = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)

    assert label == label2


def test_not_eq():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)
    label2 = GenericLabel(0, 20, document=document, a="x", y=20, z=21.0)

    assert label != label2


def test_location():
    label = GenericLabel(0, 20, document=document, a="x", y=20, z=20.0)
    assert label.location == (0, 20)


def test_get_covered_text():
    label = GenericLabel(4, 7, document=document)
    assert label.text == 'bar'


def test_covers():
    label = GenericLabel(0, 6, document=document)
    inside = GenericLabel(4, 6, document=document)
    assert label.location.covers(inside)


def test_list_attr():
    label = GenericLabel(0, 6, document=document)
    label.bar = [0, 1, 2]
    assert label.bar == [0, 1, 2]


def test_dict_attr():
    label = GenericLabel(0, 10, document=document)
    label.bar = {
        'a': 1,
        'b': 2,
    }
    assert label.bar == {
        'a': 1,
        'b': 2,
    }


def test_list_ref_loop():
    label = GenericLabel(0, 4, document=document)
    with pytest.raises(ValueError):
        label.a = [label]


def test_obj_attr():
    label = GenericLabel(0, 4, document=document)
    with pytest.raises(TypeError):
        label.b = object()


def test_float_to_int():
    label = GenericLabel(0.0, 4.0, document=document)
    assert isinstance(label.start_index, int)
    assert isinstance(label.end_index, int)


def test_get_document():
    label = GenericLabel(0, 4, document=document)
    assert label.document == document


def test_set_document():
    label = GenericLabel(0, 4)
    label.document = document
    assert label.document == document


def test_try_assign_reserved():
    l = GenericLabel(0, 0)
    with pytest.raises(ValueError):
        l.location = 10
    with pytest.raises(ValueError):
        l.text = 'blah'


def test_construct_with_reserved():
    with pytest.raises(ValueError):
        GenericLabel(0, 0, location=1)
    with pytest.raises(ValueError):
        GenericLabel(0, 0, text='some text')


def test_infinite_recursion_equals():
    a = GenericLabel(0, 15)
    b = GenericLabel(0, 13)
    a.b = b
    b.a = a

    a2 = GenericLabel(0, 15)
    b2 = GenericLabel(0, 13)
    a2.b = b2
    b2.a = a2

    assert a == a2



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mtap.processing._base import AggregateTimingInfo
from mtap.processing._base import ComponentDescriptor
from mtap.processing._base import DocumentProcessor
from mtap.processing._base import EventProcessor
from mtap.processing._base import PipelineResult
from mtap.processing._base import Processor
from mtap.processing._base import ProcessorContext
from mtap.processing._base import ProcessingComponent
from mtap.processing._base import ProcessingError
from mtap.processing._base import ProcessingResult
from mtap.processing._base import Stopwatch
from mtap.processing._base import TimerStats
from mtap.processing.descriptions import processor
from mtap.processing._pipeline import FilesInDirectoryProcessingSource
from mtap.processing._pipeline import LocalProcessor
from mtap.processing._pipeline import Pipeline
from mtap.processing._pipeline import ProcessingSource
from mtap.processing._pipeline import RemoteProcessor
from mtap.processing._service import processor_parser
from mtap.processing._service import ProcessorServer
from mtap.processing._service import run_processor



#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Module for MTAP's Data Model

Attributes:
    GENERIC_ADAPTER (~mtap.data.ProtoLabelAdapter): label adapter used for standard
        (non-distinct) :obj:`~mtap.GenericLabel`.
    DISTINCT_GENERIC_ADAPTER (~mtap.data.ProtoLabelAdapter): label adapter used for distinct
        (non-overlapping) :obj:`~mtap.GenericLabel`.
"""

from mtap.data._events import Document
from mtap.data._events import Event
from mtap.data._events import EventsClient
from mtap.data._events import Labeler
from mtap.data._base import LabelIndexType, LabelIndexInfo
from mtap.data._label_adapters import GENERIC_ADAPTER
from mtap.data._label_adapters import DISTINCT_GENERIC_ADAPTER
from mtap.data._label_adapters import ProtoLabelAdapter
from mtap.data._label_indices import label_index
from mtap.data._label_indices import presorted_label_index
from mtap.data._label_indices import LabelIndex
from mtap.data._labels import GenericLabel
from mtap.data._labels import Label
from mtap.data._labels import label
from mtap.data._labels import Location



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mtap import Event, GenericLabel, Document

from mtap.processors.copy_document import CopyDocument


def test_copy_document():
    e = Event()
    doc = Document(document_name='first', text='The quick brown fox jumped over the lazy dog.')
    e.add_document(doc)
    with doc.get_labeler('some_index') as label:
        label(0, 3, word='The')
        label(4, 9, word='quick')
        label(10, 15, word='brown')
    processor = CopyDocument('first', 'second')
    processor.process(e, {})
    second = e.documents['second']
    assert second is not None
    assert second.labels['some_index'] == [
        GenericLabel(0, 3, word='The'),
        GenericLabel(4, 9, word='quick'),
        GenericLabel(10, 15, word='brown')
    ]
